@@ -910,7 +910,7 @@ clobber_open(struct redir *f)
 	return open(unmeta(f->name), O_WRONLY | O_CREAT | O_TRUNC, 0666);
     if (stat(unmeta(f->name), &buf) || S_ISREG(buf.st_mode))
 	return open(unmeta(f->name), O_WRONLY | O_CREAT | O_EXCL, 0666);
-    if ((fd = open(unmeta(f->name), O_WRONLY | O_APPEND)) < 0)
+    if ((fd = open(unmeta(f->name), O_WRONLY)) < 0)
 	return fd;
     if (!fstat(fd, &buf) && !S_ISREG(buf.st_mode))
 	return fd;
@@ -1209,8 +1209,7 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
     }
 
     /* Do prefork substitutions */
-    prefork(args, (((type == CCASE) ? 4 : 0)
-		   | (assign ? 2 : isset(MAGICEQUALSUBST))));
+    prefork(args, assign ? 2 : isset(MAGICEQUALSUBST));
 
     if (type == SIMPLE) {
 	int unglobbed = 0;
@@ -1421,7 +1420,7 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 	    close(synch[1]);
 	    read(synch[0], &dummy, 1);
 	    close(synch[0]);
-#ifdef HAVE_DEV_FD
+#ifdef PATH_DEV_FD
 	    closem(2);
 #endif
 	    if (how & Z_ASYNC) {
@@ -1503,6 +1502,8 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 		fixfds(save);
 		execerr();
 	    }
+	    if (unset(EXECOPT))
+		continue;
 	    switch(fn->type) {
 	    case HERESTR:
 		fil = getherestr(fn);
@@ -1643,7 +1644,7 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 
 	    if (is_shfunc) {
 		/* It's a shell function */
-#ifdef HAVE_DEV_FD
+#ifdef PATH_DEV_FD
 		int i;
 
 		for (i = 10; i <= max_zsh_fd; i++)
@@ -1654,7 +1655,7 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 		    zclose(subsh_close);
 		subsh_close = -1;
 		execshfunc(cmd, (Shfunc) hn);
-#ifdef HAVE_DEV_FD
+#ifdef PATH_DEV_FD
 		for (i = 10; i <= max_zsh_fd; i++)
 		    if (fdtable[i] > 1)
 			if (--(fdtable[i]) <= 2)
@@ -1665,7 +1666,7 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 		if (forked)
 		    closem(1);
 		lastval = execbuiltin(args, (Builtin) hn);
-#ifdef HAVE_DEV_FD
+#ifdef PATH_DEV_FD
 		closem(2);
 #endif
 		if (isset(PRINTEXITVALUE) && isset(SHINSTDIN) && lastval && !subsh) {
@@ -1798,8 +1799,13 @@ restore_params(LinkList restorelist, LinkList removelist)
 
     if (removelist) {
 	/* remove temporary parameters */
-	while ((s = (char *) ugetnode(removelist)))
-	    unsetparam(s);
+	while ((s = (char *) ugetnode(removelist))) {
+	    if ((pm = (Param) paramtab->getnode(paramtab, s)) &&
+		!(pm->flags & PM_SPECIAL)) {
+		pm->flags &= ~PM_READONLY;
+		unsetparam_pm(pm, 0);
+	    }
+	}
     }
 
     if (restorelist) {
@@ -1901,7 +1907,6 @@ entersubsh(int how, int cl, int fake)
 	signal_default(SIGTTOU);
 	signal_default(SIGTTIN);
 	signal_default(SIGTSTP);
-	signal_default(SIGPIPE);
     }
     if (interact) {
 	signal_default(SIGTERM);
@@ -2074,8 +2079,8 @@ getoutput(char *cmd, int qt)
     child_unblock();
     zclose(pipes[0]);
     redup(pipes[1], 1);
+    opts[MONITOR] = 0;
     entersubsh(Z_SYNC, 1, 0);
-    signal_ignore(SIGTSTP);
     execlist(list, 0, 1);
     close(1);
     _exit(lastval);
@@ -2198,8 +2203,8 @@ getoutputfile(char *cmd)
 
     /* pid == 0 */
     redup(fd, 1);
+    opts[MONITOR] = 0;
     entersubsh(Z_SYNC, 1, 0);
-    signal_ignore(SIGTSTP);
     execlist(list, 0, 1);
     close(1);
     _exit(lastval);
@@ -2208,7 +2213,7 @@ getoutputfile(char *cmd)
     return NULL;
 }
 
-#if !defined(HAVE_DEV_FD) && defined(HAVE_FIFOS)
+#if !defined(PATH_DEV_FD) && defined(HAVE_FIFOS)
 /* get a temporary named pipe */
 
 /**/
@@ -2225,7 +2230,7 @@ namedpipe(void)
 	return NULL;
     return tnam;
 }
-#endif /* ! HAVE_DEV_FD && HAVE_FIFOS */
+#endif /* ! PATH_DEV_FD && HAVE_FIFOS */
 
 /* <(...) or >(...) */
 
@@ -2233,14 +2238,14 @@ namedpipe(void)
 char *
 getproc(char *cmd)
 {
-#if !defined(HAVE_FIFOS) && !defined(HAVE_DEV_FD)
+#if !defined(HAVE_FIFOS) && !defined(PATH_DEV_FD)
     zerr("doesn't look like your system supports FIFOs.", NULL, 0);
     return NULL;
 #else
     List list;
     int out = *cmd == Inang;
     char *pnam;
-#ifndef HAVE_DEV_FD
+#ifndef PATH_DEV_FD
     int fd;
 #else
     int pipes[2];
@@ -2248,15 +2253,15 @@ getproc(char *cmd)
 
     if (thisjob == -1)
 	return NULL;
-#ifndef HAVE_DEV_FD
+#ifndef PATH_DEV_FD
     if (!(pnam = namedpipe()))
 	return NULL;
 #else
-    pnam = ncalloc(12);
+    pnam = ncalloc(strlen(PATH_DEV_FD) + 6);
 #endif
     if (!(list = parsecmd(cmd)))
 	return NULL;
-#ifndef HAVE_DEV_FD
+#ifndef PATH_DEV_FD
     PERMALLOC {
 	if (!jobtab[thisjob].filelist)
 	    jobtab[thisjob].filelist = newlinklist();
@@ -2266,13 +2271,13 @@ getproc(char *cmd)
 #else
     mpipe(pipes);
     if (zfork()) {
-	sprintf(pnam, "/dev/fd/%d", pipes[!out]);
+	sprintf(pnam, "%s/%d", PATH_DEV_FD, pipes[!out]);
 	zclose(pipes[out]);
 	fdtable[pipes[!out]] = 2;
 #endif
 	return pnam;
     }
-#ifndef HAVE_DEV_FD
+#ifndef PATH_DEV_FD
     closem(0);
     fd = open(pnam, out ? O_WRONLY : O_RDONLY);
     if (fd == -1) {
@@ -2290,7 +2295,7 @@ getproc(char *cmd)
     zclose(out);
     _exit(lastval);
     return NULL;
-#endif   /* HAVE_FIFOS and HAVE_DEV_FD not defined */
+#endif   /* HAVE_FIFOS and PATH_DEV_FD not defined */
 }
 
 /* > >(...) or < <(...) (does not use named pipes) */
@@ -2482,11 +2487,12 @@ doshfunc(List list, LinkList doshargs, int flags, int noreturnval)
 {
     Param pm;
     char **tab, **x, *oargv0 = NULL;
-    int xexittr, oldzoptind, oldlastval;
+    int xexittr, newexittr, oldzoptind, oldlastval;
     LinkList olist;
     char *s, *ou;
-    void *xexitfn;
+    void *xexitfn, *newexitfn;
     char saveopts[OPT_SIZE];
+    int obreaks = breaks;
 
     HEAPALLOC {
 	pushheap();
@@ -2552,7 +2558,10 @@ doshfunc(List list, LinkList doshargs, int flags, int noreturnval)
 	zfree(locallist, sizeof(struct linklist));
 
 	locallist = olist;	/* restore the old list of local variables */
-	breaks = retflag = 0;
+	if (retflag) {
+	    retflag = 0;
+	    breaks = obreaks;
+	}
 	freearray(pparams);
 	if (oargv0) {
 	    zsfree(argzero);
@@ -2572,11 +2581,15 @@ doshfunc(List list, LinkList doshargs, int flags, int noreturnval)
 	    opts[LOCALOPTIONS] = saveopts[LOCALOPTIONS];
 	}
 
-	/* Did the shell function define a trap on SIGEXIT? */
-	if (sigfuncs[SIGEXIT]) {
-	    dotrap(SIGEXIT);
-	    unsettrap(SIGEXIT);
-	}
+	/*
+	 * The trap '...' EXIT runs in the environment of the caller,
+	 * so remember it here but run it after resetting the
+	 * traps for the parent.
+	 */
+	newexittr = sigtrapped[SIGEXIT];
+	newexitfn = sigfuncs[SIGEXIT];
+	if (newexittr & ZSIG_FUNC)
+	    shfunctab->removenode(shfunctab, "TRAPEXIT");
 
 	sigtrapped[SIGEXIT] = xexittr;
 	if (xexittr & ZSIG_FUNC) {
@@ -2584,6 +2597,12 @@ doshfunc(List list, LinkList doshargs, int flags, int noreturnval)
 	    sigfuncs[SIGEXIT] = ((Shfunc) xexitfn)->funcdef;
 	} else
 	    sigfuncs[SIGEXIT] = (List) xexitfn;
+
+	if (newexitfn) {
+	    dotrapargs(SIGEXIT, &newexittr, newexitfn);
+	    freestruct(newexitfn);
+	}
+
 	if (trapreturn < -1)
 	    trapreturn++;
 	if (noreturnval)
