@@ -1,5 +1,5 @@
 /*
- * $Id: glob.c,v 2.31 1996/07/27 20:24:36 hzoli Exp $
+ * $Id: glob.c,v 2.35 1996/10/15 20:16:35 hzoli Exp $
  *
  * glob.c - filename generation
  *
@@ -10,9 +10,9 @@
  *
  * Permission is hereby granted, without written agreement and without
  * license or royalty fees, to use, copy, modify, and distribute this
- * software and its documentation for any purpose, provided that the
- * above copyright notice and the following two paragraphs appear in
- * all copies of this software.
+ * software and to distribute modified versions of this software for any
+ * purpose, provided that the above copyright notice and the following
+ * two paragraphs appear in all copies of this software.
  *
  * In no event shall Paul Falstad or the Zsh Development Group be liable
  * to any party for direct, indirect, special, incidental, or consequential
@@ -187,6 +187,7 @@ glob(LinkList list, LinkNode np)
 		} else if (*s == ',') {
 		    /* A comma separates alternative sets of qualifiers */
 		    s++;
+		    sense = 0;
 		    if (qualct) {
 			qn = (struct qual *)hcalloc(sizeof *qn);
 			qo->or = qn;
@@ -496,16 +497,6 @@ glob(LinkList list, LinkNode np)
 		    return;
 	    }
 	}
-    } else if ((str[sl - 1] == '/') &&
-	       !((str[sl - 2] == Star) &&
-		 (str[sl - 3] == Star))) {	/* foo/ == foo(-/) */
-	/* No explicit qualifiers, but implicitly match final directory */
-	str[sl - 1] = '\0';
-	quals = (struct qual *)hcalloc(sizeof *quals);
-	quals->func = qualmode;
-	quals->data = S_IFDIR;
-	quals->sense = 2;
-	qualct = 1;
     }
     if (*str == '/') {		/* pattern has absolute path */
 	str++;
@@ -780,7 +771,7 @@ haswilds(char *str)
 		mod = NULL;
 	    break;
 	case ':':
-	    if (parlev == 1)
+	    if (parlev == 1 && !mod)
 		mod = str;
 	    break;
 	case Bar:
@@ -788,13 +779,15 @@ haswilds(char *str)
 		*str = '|';
 		break;
 	    } /* else fall through */
-	case Pound:
-	case Hat:
 	case Star:
 	case Inbrack:
 	case Inang:
 	case Quest:
 	    return str;
+	case Pound:
+	case Hat:
+	    if (isset(EXTENDEDGLOB))
+		return str;
 	}
     if (!mod || parlev)
 	return NULL;
@@ -809,57 +802,88 @@ haswilds(char *str)
 int
 hasbraces(char *str)
 {
-    int mb, bc, cmct1, cmct2;
-    char *lbr = NULL;
+    char *lbr, *mbr, *comma;
 
     if (isset(BRACECCL)) {
 	/* In this case, any properly formed brace expression  *
 	 * will match and expand to the characters in between. */
-	for (mb = bc = 0; *str; ++str)
+	int bc;
+
+	for (bc = 0; *str; ++str)
 	    if (*str == Inbrace) {
-		if (!mb && str[1] == Outbrace)
+		if (!bc && str[1] == Outbrace)
 		    *str++ = '{', *str = '}';
-		else if (++bc > mb)
-		    mb = bc;
+		else
+		    bc++;
 	    } else if (*str == Outbrace)
-		if (--bc < 0)
-		    return (0);
-	return (mb && bc == 0);
+		if (!bc)
+		    *str = '}';
+		else if (!--bc)
+		    return 1;
+	return 0;
     }
     /* Otherwise we need to look for... */
-    for (mb = bc = cmct1 = cmct2 = 0; *str; str++) {
-	if (*str == Inbrace) {
-	    if (!bc)
-		lbr = str;
-	    bc++;
-	    if (str[2] == '-' && str[3] && str[4] == Outbrace) { /* {a-z} */
-		/* ...character ranges... */
-		cmct1++;
-		if (bc == 1)
-		    cmct2++;
-	    }
-	} else if (*str == Outbrace) {
-	    /* (see if we've finished this set of braces) */
-	    bc--;
-	    if (!bc) {
-		if (!cmct2) {
-		    *lbr = '{';
-		    *str = '}';
+    lbr = mbr = comma = NULL;
+    for (;;) {
+	switch (*str++) {
+	case Inbrace:
+	    if (!lbr) {
+		lbr = str - 1;
+		while (idigit(*str))
+		    str++;
+		if (*str == '.' && str[1] == '.') {
+		    str++;
+		    while (idigit(*++str));
+		    if (*str == Outbrace &&
+			(idigit(lbr[1]) || idigit(str[-1])))
+			return 1;
 		}
-		cmct2 = 0;
+	    } else {
+		char *s = --str;
+
+		if (skipparens(Inbrace, Outbrace, &str)) {
+		    *lbr = *s = '{';
+		    if (comma)
+			str = comma;
+		    if (mbr && mbr < str)
+			str = mbr;
+		    lbr = mbr = comma = NULL;
+		} else if (!mbr)
+		    mbr = s;
 	    }
-	} else if ((*str == Comma || (*str == '.' && str[1] == '.')) && bc) {
-	    /* ...or comma'd ranges or numeric ranges. */
-	    cmct1++;
-	    if (bc == 1)
-		cmct2++;
+	    break;
+	case Outbrace:
+	    if (!lbr)
+		str[-1] = '}';
+	    else if (comma)
+		return 1;
+	    else {
+		*lbr = '{';
+		str[-1] = '}';
+		if (mbr)
+		    str = mbr;
+		mbr = lbr = NULL;
+	    }
+	    break;
+	case Comma:
+	    if (!lbr)
+		str[-1] = ',';
+	    else if (!comma)
+		comma = str - 1;
+	    break;
+	case '\0':
+	    if (lbr)
+		*lbr = '{';
+	    if (!mbr && !comma)
+		return 0;
+	    if (comma)
+		str = comma;
+	    if (mbr && mbr < str)
+		str = mbr;
+	    lbr = mbr = comma = NULL;
+	    break;
 	}
-	if (bc > mb)
-	    mb = bc;
-	if (bc < 0)
-	    return 0;
     }
-    return (mb && bc == 0 && cmct1);
 }
 
 /* expand stuff like >>*.c */
@@ -977,7 +1001,8 @@ xpandbraces(LinkList list, LinkNode *np)
 		++comma;	/* we have {foo,bar} */
 	    else if (*str2 == '.' && str2[1] == '.')
 		dotdot++;	/* we have {num1..num2} */
-    if (!comma && !bc && dotdot) {
+    DPUTS(bc, "BUG: unmatched brace in xpandbraces()");
+    if (!comma && dotdot) {
 	/* Expand range like 0..10 numerically: comma or recursive
 	   brace expansion take precedence. */
 	char *dots, *p;
@@ -1021,21 +1046,26 @@ xpandbraces(LinkList list, LinkNode *np)
 	    return;
 	}
     }
-    if (!comma && !bc && isset(BRACECCL)) {	/* {a-mnop} */
+    if (!comma && isset(BRACECCL)) {	/* {a-mnop} */
 	/* Here we expand each character to a separate node,      *
 	 * but also ranges of characters like a-m.  ccl is a      *
 	 * set of flags saying whether each character is present; *
 	 * the final list is in lexical order.                    */
 	char ccl[256], *p;
 	unsigned char c1, c2, lastch;
+	unsigned int len, pl;
 
 	uremnode(list, node);
 	memset(ccl, 0, sizeof(ccl) / sizeof(ccl[0]));
 	for (p = str + 1, lastch = 0; p < str2;) {
 	    if (itok(c1 = *p++))
 		c1 = ztokens[c1 - STOUC(Pound)];
+	    if ((char) c1 == Meta)
+		c1 = 32 ^ *p++;
 	    if (itok(c2 = *p))
 		c2 = ztokens[c2 - STOUC(Pound)];
+	    if ((char) c2 == Meta)
+		c2 = 32 ^ p[1];
 	    if (c1 == '-' && lastch && p < str2 && (int)lastch <= (int)c2) {
 		while ((int)lastch < (int)c2)
 		    ccl[lastch++] = 1;
@@ -1043,50 +1073,29 @@ xpandbraces(LinkList list, LinkNode *np)
 	    } else
 		ccl[lastch = c1] = 1;
 	}
-	strcpy(str + 1, str2 + 1);
+	pl = str - str3;
+	len = pl + strlen(++str2) + 2;
 	for (p = ccl + 255; p-- > ccl;)
 	    if (*p) {
-		*str = p - ccl;
-		insertlinknode(list, last, dupstring(str3));
+		c1 = p - ccl;
+		if (imeta(c1)) {
+		    str = ncalloc(len + 1);
+		    str[pl] = Meta;
+		    str[pl+1] = c1 ^ 32;
+		    strcpy(str + pl + 2, str2);
+		} else {
+		    str = ncalloc(len);
+		    str[pl] = c1;
+		    strcpy(str + pl + 1, str2);
+		}
+		memcpy(str, str3, pl);
+		insertlinknode(list, last, str);
 	    }
 	*np = nextnode(last);
 	return;
     }
-    if (str[2] == '-' && str[3] && str[4] == Outbrace) { /* {a-z} */
-	/* Now any other ranges present: note this only happens       *
-         * for a pattern like {<char>-<char>}, but it happens for ANY *
-         * character <char>, so {,-z} does this (possibly a bug).     */
-	char c1, c2;
-
-	uremnode(list, node);
-	chuck(str);
-	c1 = *str;
-	chuck(str);
-	chuck(str);
-	c2 = *str;
-	chuck(str);
-	if (itok(c1))
-	    c1 = ztokens[c1 - Pound];
-	if (itok(c2))
-	    c2 = ztokens[c2 - Pound];
-	if (c1 < c2)
-	    for (; c2 >= c1; c2--) {	/* {a-z} */
-		*str = c2;
-		insertlinknode(list, last, dupstring(str3));
-	} else
-	    for (; c2 <= c1; c2++) {	/* {z-a} */
-		*str = c2;
-		insertlinknode(list, last, dupstring(str3));
-	    }
-	*np = nextnode(last);
-	return;
-    }
-    prev = str - str3;
-    str2 = getparen(str++);
-    if (!str2) {
-	zerr("how did you get this error?", NULL, 0);
-	return;
-    }
+    prev = str++ - str3;
+    str2++;
     uremnode(list, node);
     node = last;
     /* Finally, normal comma expansion               *
@@ -1097,13 +1106,13 @@ xpandbraces(LinkList list, LinkNode *np)
 	int cnt;
 
 	for (str4 = str, cnt = 0; cnt || (*str != Comma && *str !=
-					  Outbrace); str++)
+					  Outbrace); str++) {
 	    if (*str == Inbrace)
 		cnt++;
 	    else if (*str == Outbrace)
 		cnt--;
-	    else if (!*str)
-		exit(10);	/* slightly overemphatic, perhaps??? */
+	    DPUTS(!*str, "BUG: illegal brace expansion");
+	}
 	/* Concatenate the string before the braces (str3), the section *
 	 * just found (str4) and the text after the braces (str2)       */
 	zz = (char *)ncalloc(prev + (str - str4) + strlen(str2) + 1);
@@ -1121,25 +1130,6 @@ xpandbraces(LinkList list, LinkNode *np)
     *np = nextnode(last);
 }
 
-/* get closing paren, given pointer to opening paren */
-
-/**/
-char *
-getparen(char *str)
-{
-    int cnt = 1;
-    char typein = *str++, typeout = typein + 1;
-
-    for (; *str && cnt; str++)
-	if (*str == typein)
-	    cnt++;
-	else if (*str == typeout)
-	    cnt--;
-    if (!str && cnt)
-	return NULL;
-    return str;
-}
-
 /* check to see if a matches b (b is not a filename pattern) */
 
 /**/
@@ -1532,17 +1522,14 @@ scanner(Complist q)
 	    /* Do pattern matching on current path section. */
 	    char *fn;
 	    int dirs = !!q->next;
-	    struct dirent *de;
 	    DIR *lock = opendir((*pathbuf) ? unmeta(pathbuf) : ".");
 
 	    if (lock == NULL)
 		return;
-	    while ((de = zreaddir(lock))) {
+	    while ((fn = zreaddir(lock))) {
 		/* Loop through the directory */
 		if (errflag)
 		    break;
-		/* fn is current file name in directory */
-		fn = &de->d_name[0];
 		/* skip this and parent directory */
 		if (fn[0] == '.'
 		    && (fn[1] == '\0'
@@ -1791,13 +1778,12 @@ doesmatch(Comp c)
 	    unsigned long t1, t2, t3;
 	    char *ptr;
 
+	    if (!idigit(*pptr))
+		break;
 	    if (*++pat == Outang || 
 		(*pat == '-' && pat[1] == Outang && ++pat)) {
 		/* <> or <->:  any number matches */
-		(void)zstrtol(pptr, &ptr, 10);
-		if (ptr == pptr)
-		    break;
-		pptr = ptr;
+		while (idigit(*++pptr));
 		pat++;
 	    } else {
 		/* Flag that there is no upper limit */
@@ -1808,18 +1794,19 @@ doesmatch(Comp c)
 		 * unsigned arithmetic.
 		 */
 		t1 = (unsigned long)zstrtol(pptr, &ptr, 10);
-		if (ptr == pptr)
-		    break;
 		pptr = ptr;
 		/* t2 = lower limit */
-		t2 = (unsigned long)zstrtol(pat, &ptr, 10);
+		if (idigit(*pat))
+		    t2 = (unsigned long)zstrtol(pat, &ptr, 10);
+		else
+		    t2 = 0, ptr = pat;
 		if (*ptr != '-' || (not3 = (ptr[1] == Outang)))
 				/* exact match or no upper limit */
 		    t3 = t2, pat = ptr + not3;
 		else		/* t3 = upper limit */
 		    t3 = (unsigned long)zstrtol(ptr + 1, &pat, 10);
-		if (*pat++ != Outang)
-		    exit(21);
+		DPUTS(*pat != Outang, "BUG: wrong internal range pattern");
+		pat++;
 		if (t1 < t2 || (!not3 && t1 > t3))
 		    break;
 	    }
@@ -1897,7 +1884,7 @@ parsecomplist(void)
 
     /* Parse repeated directories such as (dir/)# and (dir/)## */
     if (*(str = pptr) == Inpar && !skipparens(Inpar, Outpar, &str) &&
-        *str == Pound && str[-2] == '/') {
+        *str == Pound && isset(EXTENDEDGLOB) && str[-2] == '/') {
 	pptr++;
 	if (!(c1 = parsecompsw(0)))
 	    return NULL;
@@ -1942,10 +1929,11 @@ Comp
 parsecomp(void)
 {
     Comp c = (Comp) alloc(sizeof *c), c1, c2;
-    char cstr[PATH_MAX * 2], *s = cstr, *ls = NULL;
+    char *cstr, *ls = NULL;
 
     /* In case of alternatives, code coming up is stored in tail. */
     c->next = tail;
+    cstr = pptr;
 
     while (*pptr && (mode || *pptr != '/') && *pptr != Bar &&
 	   (unset(EXTENDEDGLOB) || *pptr != Tilde ||
@@ -1954,14 +1942,12 @@ parsecomp(void)
 	/* Go through code until we find something separating alternatives,
 	 * or path components if relevant.
 	 */
-	if (*pptr == Hat) {
+	if (*pptr == Hat && isset(EXTENDEDGLOB)) {
 	    /* negate remaining pattern */
-	    *s++ = Hat;
-	    *s++ = '\0';
 	    pptr++;
+	    c->str = dupstrpfx(cstr, pptr - cstr);
 	    if (!(c->next = parsecomp()))
 		return NULL;
-	    c->str = dupstring(cstr);
 	    return c;
 	}
 	if (*pptr == Star && pptr[1] &&
@@ -1972,7 +1958,7 @@ parsecomp(void)
 	     * (zero or more repetitions) of the single character pattern
 	     * operator `?'.
 	     */
-	    *s++ = '\0';
+	    c->str = dupstrpfx(cstr, pptr - cstr);
 	    pptr++;
 	    c1 = (Comp) alloc(sizeof *c1);
 	    *(c1->str = dupstring("?")) = Quest;
@@ -1981,7 +1967,6 @@ parsecomp(void)
 		return NULL;
 	    c1->next = c2;
 	    c->next = c1;
-	    c->str = dupstring(cstr);
 	    return c;
 	}
 	if (*pptr == Inpar) {
@@ -1995,7 +1980,7 @@ parsecomp(void)
 		errflag = 1;
 		return NULL;
 	    }
-	    if (*pptr == Pound) {
+	    if (*pptr == Pound && isset(EXTENDEDGLOB)) {
 		/* Zero (or one) or more repetitions of group */
 		dpnd = 1;
 		pptr++;
@@ -2012,75 +1997,71 @@ parsecomp(void)
 	    /* ...before going back and parsing inside the group. */
 	    endp = pptr;
 	    pptr = startp;
+	    c->str = dupstrpfx(cstr, pptr - cstr);
 	    pptr++;
-	    *s++ = '\0';
 	    c->next = (Comp) alloc(sizeof *c);
-	    c->next->left = parsecompsw(0);
+	    if (!(c->next->left = parsecompsw(0)))
+		return NULL;
 	    /* Remember closures for group. */
 	    if (dpnd)
 		c->next->stat |= (dpnd == 2) ? C_TWOHASH : C_ONEHASH;
 	    c->next->next = dpnd ? c1 : (Comp) alloc(sizeof *c);
 	    pptr = endp;
 	    tail = stail;
-	    c->str = dupstring(cstr);
 	    return c;
 	}
-	if (*pptr == Pound) {
+	if (*pptr == Pound && isset(EXTENDEDGLOB)) {
 	    /* repeat whatever we've just had (ls) zero or more times */
-	    *s = '\0';
-	    pptr++;
 	    if (!ls)
 		return NULL;
+	    c2 = (Comp) alloc(sizeof *c);
+	    c2->str = dupstrpfx(ls, pptr - ls);
+	    pptr++;
 	    if (*pptr == Pound) {
 		/* need one or more matches: cheat by copying previous char */
 		pptr++;
 		c->next = c1 = (Comp) alloc(sizeof *c);
-		c1->str = dupstring(ls);
+		c1->str = c2->str;
 	    } else
 		c1 = c;
-	    c1->next = c2 = (Comp) alloc(sizeof *c);
-	    c2->str = dupstring(ls);
+	    c1->next = c2;
 	    c2->stat |= C_ONEHASH;
 	    /* parse the rest of the pattern and return. */
 	    c2->next = parsecomp();
 	    if (!c2->next)
 		return NULL;
-	    *ls++ = '\0';
-	    c->str = dupstring(cstr);
+	    c->str = dupstrpfx(cstr, ls - cstr);
 	    return c;
 	}
-	ls = s;			/* whatever we just parsed */
+	ls = pptr;		/* whatever we just parsed */
 	if (*pptr == Inang) {
 	    /* Numeric glob */
 	    int dshct;
 
 	    dshct = (pptr[1] == Outang);
-	    *s++ = *pptr++;
-	    while (*pptr && (*s++ = *pptr++) != Outang)
-		if (s[-1] == '-')
-		    dshct++;
-		else if (!idigit(s[-1]))
+	    while (*++pptr && *pptr != Outang)
+		if (*pptr == '-' && !dshct)
+		    dshct = 1;
+		else if (!idigit(*pptr))
 		    break;
-	    if (s[-1] != Outang)
+	    if (*pptr != Outang)
 		return NULL;
 	} else if (*pptr == Inbrack) {
 	    /* Character set: brackets had better match */
-	    while (*pptr && (*s++ = *pptr++) != Outbrack);
-	    if (s[-1] != Outbrack)
+	    while (*++pptr && *pptr != Outbrack)
+		if (itok(*pptr))
+		    *pptr = ztokens[*pptr - Pound];
+	    if (*pptr != Outbrack)
 		return NULL;
-	} else if (itok(*pptr) && *pptr != Star && *pptr != Quest) {
+	} else if (itok(*pptr) && *pptr != Star && *pptr != Quest)
 	    /* something that can be tokenised which isn't otherwise special */
-	    *s++ = ztokens[*pptr++ - Pound];
-	} else {
-	    /* any other character */
-	    *s++ = *pptr++;
-	}
+	    *pptr = ztokens[*pptr - Pound];
+	pptr++;
     }
     /* mark if last pattern component in path component or pattern */
     if (*pptr == '/' || !*pptr)
 	c->stat |= C_LAST;
-    *s++ = '\0';
-    c->str = dupstring(cstr);
+    c->str = dupstrpfx(cstr, pptr - cstr);
     return c;
 }
 
