@@ -204,6 +204,10 @@ EXTERN int inbufct;
 
 EXTERN int inbufflags;
 
+/* flag that an alias should be expanded after expansion ending in space */
+
+EXTERN int inalmore;
+
 /* != 0 if this is a subshell */
  
 EXTERN int subsh;
@@ -447,9 +451,9 @@ EXTERN int spaceflag;
  
 EXTERN int nocorrect;
 
-/* != 0 means we have removed the current event from the history List */
+/* state of the history mechanism (see hist.c) */
  
-EXTERN int histremmed;
+EXTERN int histactive;
 
 /* current emulation (used to decide which set of option letters is used) */
 
@@ -495,9 +499,9 @@ EXTERN unsigned txtchange;
 
 EXTERN char *term;		/* $TERM */
  
-/* != 0 if this $TERM setup is usable */
- 
-EXTERN int termok;
+/* 0 if this $TERM setup is usable, otherwise it contains TERM_* flags */
+
+EXTERN int termflags;
  
 /* flag for CSHNULLGLOB */
  
@@ -599,6 +603,10 @@ EXTERN struct list dummy_list;
 EXTERN int tclen[TC_COUNT];
  
 EXTERN char *tcstr[TC_COUNT];
+
+/* Values of the li and co entries */
+
+EXTERN int tclines, tccolumns;
  
 /* names of the strings we want */
 #ifdef GLOBALS
@@ -748,7 +756,7 @@ struct option optns[OPT_SIZE] = {
     {"monitor", 		'm',  'm',  OPT_SPECIAL},
     {"multios", 		0,    0,    OPT_EMULATE|OPT_ZSH},
     {"nomatch", 		x'3', 0,    OPT_EMULATE|OPT_NONBOURNE},
-    {"notify", 			'5',  'b',  OPT_ALL},
+    {"notify", 			'5',  'b',  OPT_ZSH},
     {"nullglob", 		'G',  0,    OPT_EMULATE},
     {"numericglobsort", 	0,    0,    0},
     {"overstrike", 		0,    0,    0},
