@@ -1,5 +1,5 @@
 /*
- * $Id: hashtable.c,v 2.10 1996/07/26 12:54:36 hzoli Exp $
+ * $Id: hashtable.c,v 2.11 1996/10/15 20:16:35 hzoli Exp $
  *
  * hashtable.c - hash tables
  *
@@ -10,9 +10,9 @@
  *
  * Permission is hereby granted, without written agreement and without
  * license or royalty fees, to use, copy, modify, and distribute this
- * software and its documentation for any purpose, provided that the
- * above copyright notice and the following two paragraphs appear in
- * all copies of this software.
+ * software and to distribute modified versions of this software for any
+ * purpose, provided that the above copyright notice and the following
+ * two paragraphs appear in all copies of this software.
  *
  * In no event shall Paul Falstad or the Zsh Development Group be liable
  * to any party for direct, indirect, special, incidental, or consequential
@@ -541,19 +541,22 @@ hashdir(char **dirp)
 {
     Cmdnam cn;
     DIR *dir;
-    struct dirent *de;
+    char *fn;
 
     if (isrelative(*dirp) || !(dir = opendir(unmeta(*dirp))))
 	return;
 
-    readdir(dir);
-    readdir(dir);
-    while ((de = zreaddir(dir))) {
-	if (!cmdnamtab->getnode(cmdnamtab, de->d_name)) {
+    while ((fn = zreaddir(dir))) {
+	/* Ignore `.' and `..'. */
+	if (fn[0] == '.' &&
+	    (fn[1] == '\0' ||
+	     (fn[1] == '.' && fn[2] == '\0')))
+	    continue;
+	if (!cmdnamtab->getnode(cmdnamtab, fn)) {
 	    cn = (Cmdnam) zcalloc(sizeof *cn);
 	    cn->flags = 0;
 	    cn->u.name = dirp;
-	    cmdnamtab->addnode(cmdnamtab, ztrdup(de->d_name), cn);
+	    cmdnamtab->addnode(cmdnamtab, ztrdup(fn), cn);
 	}
     }
     closedir(dir);
@@ -694,8 +697,12 @@ void
 disableshfuncnode(HashNode hn, int flags)
 {
     hn->flags |= DISABLED;
-    if (!strncmp(hn->nam, "TRAP", 4))
-	unsettrap(getsignum(hn->nam + 4));
+    if (!strncmp(hn->nam, "TRAP", 4)) {
+	int signum = getsignum(hn->nam + 4);
+	sigtrapped[signum] &= ~ZSIG_FUNC;
+	sigfuncs[signum] = NULL;
+	unsettrap(signum);
+    }
 }
 
 /* Re-enable an entry in the shell function hash table.  *
