@@ -398,8 +398,10 @@ init_term(void)
     static char termbuf[2048];	/* the termcap buffer */
 #endif
 
-    if (!*term)
-	return termok = TERM_BAD;
+    if (!*term) {
+	termflags |= TERM_UNKNOWN;
+	return 0;
+    }
 
     /* unset zle if using zsh under emacs */
     if (!strcmp(term, "emacs"))
@@ -415,12 +417,14 @@ init_term(void)
 	if (isset(INTERACTIVE))
 	    zerr("can't find termcap info for %s", term, 0);
 	errflag = 0;
-	return termok = TERM_BAD;
+	termflags |= TERM_BAD;
+	return 0;
     } else {
 	char tbuf[1024], *pp;
 	int t0;
 
-	termok = TERM_OK;
+	termflags &= ~TERM_BAD;
+	termflags &= ~TERM_UNKNOWN;
 	for (t0 = 0; t0 != TC_COUNT; t0++) {
 	    pp = tbuf;
 	    zsfree(tcstr[t0]);
@@ -437,12 +441,17 @@ init_term(void)
 	/* check whether terminal has automargin (wraparound) capability */
 	hasam = tgetflag("am");
 
+	tclines = tgetnum("li");
+	tccolumns = tgetnum("co");
+
 	/* if there's no termcap entry for cursor up, use single line mode: *
-	 * this is flagged by termok which is examined in zle_refresh.c     *
+	 * this is flagged by termflags which is examined in zle_refresh.c  *
 	 */
-	if (!tccan(TCUP)) {
-		tcstr[TCUP] = NULL;
-		termok = TERM_NOUP;
+	if (tccan(TCUP))
+	    termflags &= ~TERM_NOUP;
+	else {
+	    tcstr[TCUP] = NULL;
+	    termflags |= TERM_NOUP;
 	}
 
 	/* if there's no termcap entry for cursor left, use \b. */
@@ -464,7 +473,7 @@ init_term(void)
 	    tclen[TCCLEARSCREEN] = 1;
 	}
     }
-    return termok;
+    return 1;
 }
 
 /* Initialize lots of global variables and hash tables */
@@ -492,7 +501,7 @@ setupvals(void)
     bangchar = '!';
     hashchar = '#';
     hatchar = '^';
-    termok = TERM_BAD;
+    termflags = TERM_UNKNOWN;
     curjob = prevjob = coprocin = coprocout = -1;
     gettimeofday(&shtimer, &dummy_tz);	/* init $SECONDS */
     srand((unsigned int)(shtimer.tv_sec + shtimer.tv_usec)); /* seed $RANDOM */
@@ -541,20 +550,6 @@ setupvals(void)
     mypid = (long) getpid();
     term  = ztrdup("");
 
-#ifdef TIOCGWINSZ
-    if (!(columns = shttyinfo.winsize.ws_col))
-	columns = 80;
-    if (columns < 2)
-	opts[USEZLE] = 0;
-    if (!(lines = shttyinfo.winsize.ws_row))
-	lines = 24;
-    if (lines < 2)
-	opts[SINGLELINEZLE] = 1;
-#else
-    columns = 80;
-    lines = 24;
-#endif
-
     /* The following variable assignments cause zsh to behave more *
      * like Bourne and Korn shells when invoked as "sh" or "ksh".  *
      * NULLCMD=":" and READNULLCMD=":"                             */
@@ -603,6 +598,14 @@ setupvals(void)
     createnameddirtable();  /* create hash table for named directories */
     createparamtable();     /* create paramater hash table             */
 
+#ifdef TIOCGWINSZ
+    adjustwinsize();
+#else
+    /* Using zero below sets the defaults from termcap */
+    setiparam("COLUMNS", 0);
+    setiparam("LINES", 0);
+#endif
+
     /* create hash table for multi-character emacs bindings */
     createemkeybindtable();
 
@@ -624,7 +627,7 @@ setupvals(void)
     locallist = NULL;
     locallevel = sourcelevel = 0;
     trapreturn = 0;
-    noerrexit = 0;
+    noerrexit = -1;
     nohistsave = 1;
     dirstack = newlinklist();
     bufstack = newlinklist();
@@ -656,11 +659,11 @@ init_signals(void)
 
     install_handler(SIGHUP);
     install_handler(SIGCHLD);
+#ifdef SIGWINCH
+    install_handler(SIGWINCH);
+#endif
     if (interact) {
 	install_handler(SIGALRM);
-#ifdef SIGWINCH
-	install_handler(SIGWINCH);
-#endif
 	signal_ignore(SIGTERM);
     }
     if (jobbing) {
@@ -674,7 +677,6 @@ init_signals(void)
 	    signal_ignore(SIGTTOU);
 	    signal_ignore(SIGTSTP);
 	    signal_ignore(SIGTTIN);
-	    signal_ignore(SIGPIPE);
 	    attachtty(mypgrp);
 	}
     }
@@ -768,7 +770,6 @@ init_misc(void)
 
     if (interact && isset(RCS))
 	readhistfile(getsparam("HISTFILE"), 0);
-    adjustwinsize();   /* check window size and adjust if necessary */
 }
 
 /* source a file */
