@@ -58,10 +58,13 @@
  * inpop(), which effectively flushes any unread input as well as restoring
  * the previous input state.
  *
- * The flags INP_ALMORE is set when a popped alias ends in a space so that
- * the next word should also be alias expanded.  The flag is reset in
- * exalias().  The internal flag INP_ALCONT shows that the stack element
- * was pushed by an alias expansion and should not be needed elsewhere.
+ * The internal flag INP_ALCONT shows that the stack element was pushed
+ * by an alias expansion; it should not be needed elsewhere.
+ *
+ * The global variable inalmore is set to indicate aliases should
+ * continue to be expanded because the last alias expansion ended
+ * in a space.  It is only reset after a complete word was read
+ * without expanding a new alias, in exalias().
  *
  * PWS 1996/12/10
  */
@@ -420,7 +423,7 @@ inpush(char *str, int flags, Alias inalias)
     instacktop->bufptr = inbufptr;
     instacktop->bufleft = inbufleft;
     instacktop->bufct = inbufct;
-    inbufflags &= ~(INP_ALCONT|INP_ALMORE);
+    inbufflags &= ~INP_ALCONT;
     if (flags & (INP_ALIAS|INP_HIST)) {
 	/*
 	 * Text is expansion for history or alias, so continue
@@ -486,7 +489,7 @@ inpoptop(void)
 	/* a real alias:  mark it as unused. */
 	instacktop->alias->inuse = 0;
 	if (*t && t[strlen(t) - 1] == ' ') {
-	    inbufflags |= INP_ALMORE;
+	    inalmore = 1;
 	    histbackword();
 	}
     }
