@@ -1,5 +1,5 @@
 /*
- * $Id: jobs.c,v 2.14 1996/07/31 18:15:35 hzoli Exp $
+ * $Id: jobs.c,v 2.17 1996/10/15 20:16:35 hzoli Exp $
  *
  * jobs.c - job control
  *
@@ -10,9 +10,9 @@
  *
  * Permission is hereby granted, without written agreement and without
  * license or royalty fees, to use, copy, modify, and distribute this
- * software and its documentation for any purpose, provided that the
- * above copyright notice and the following two paragraphs appear in
- * all copies of this software.
+ * software and to distribute modified versions of this software for any
+ * purpose, provided that the above copyright notice and the following
+ * two paragraphs appear in all copies of this software.
  *
  * In no event shall Paul Falstad or the Zsh Development Group be liable
  * to any party for direct, indirect, special, incidental, or consequential
@@ -121,7 +121,6 @@ update_job(Job jn)
     int job;
     int val = 0, status = 0;
     int somestopped = 0, inforeground = 0;
-    pid_t pgrp;
 
     for (pn = jn->procs; pn; pn = pn->next) {
 	if (pn->status == SP_RUNNING)      /* some processes in this job are running       */
@@ -155,13 +154,20 @@ update_job(Job jn)
 	}
     }
 
-    pgrp = gettygrp();           /* get process group of tty      */
-
-    /* is this job in the foreground of an interactive shell? */
-    if ((jn->gleader == pgrp || (pgrp > 1 && kill(-pgrp, 0) == -1)) &&
-	!ttyfrozen && !val && !jn->stty_in_env)
+    if (shout && !ttyfrozen && !jn->stty_in_env)
 	gettyinfo(&shttyinfo);
-    adjustwinsize();             /* check window size and adjust if necessary */
+
+    if (isset(MONITOR)) {
+	pid_t pgrp = gettygrp();           /* get process group of tty      */
+
+	/* is this job in the foreground of an interactive shell? */
+	if (mypgrp != pgrp && inforeground &&
+	    (jn->gleader == pgrp || (pgrp > 1 && kill(-pgrp, 0) == -1))) {
+	    attachtty(mypgrp);
+	    adjustwinsize();   /* check window size and adjust if necessary */
+	}
+    }
+
     if (somestopped && jn->stat & STAT_SUPERJOB)
 	return;
     jn->stat |= (somestopped) ? STAT_CHANGED | STAT_STOPPED :
@@ -178,30 +184,29 @@ update_job(Job jn)
     if (sigtrapped[SIGCHLD] && job != thisjob)
 	dotrap(SIGCHLD);
 
-    /* WHY DO WE USE THE RETURN STATUS OF PROCESS GROUP LEADER HERE? */
-    /* If the foreground job got a signal, pretend we got it, too.   */
-    if (inforeground && WIFSIGNALED(status)) {
-	if (sigtrapped[WTERMSIG(status)]) {
-	    /* Run the trap with the error flag unset.
-	     * Errflag is set in printjobs if the jobs terminated
-	     * with SIGINT.  I don't know why it's done there and
-	     * not here.   (PWS 1995/06/08)
-	     */
-	    errflag = 0;
-	    dotrap(WTERMSIG(status));
-	    /* We keep the errflag as set or not by dotrap.
-	     * This is to fulfil the promise to carry on
-	     * with the jobs if trap returns zero.
-	     * Setting breaks = loops ensures a consistent return
-	     * status if inside a loop.  Maybe the code in loops
-	     * should be changed.
-	     */
-	    if (errflag)
+    /* When MONITOR is set, the foreground process runs in a different *
+     * process group from the shell, so the shell will not receive     *
+     * terminal signals, therefore we we pretend that the shell got    *
+     * the signal too.                                                 */
+    if (inforeground && isset(MONITOR) && WIFSIGNALED(status)) {
+	int sig = WTERMSIG(status);
+
+	if (sig == SIGINT || sig == SIGQUIT) {
+	    if (sigtrapped[sig]) {
+		dotrap(sig);
+		/* We keep the errflag as set or not by dotrap.
+		 * This is to fulfil the promise to carry on
+		 * with the jobs if trap returns zero.
+		 * Setting breaks = loops ensures a consistent return
+		 * status if inside a loop.  Maybe the code in loops
+		 * should be changed.
+		 */
+		if (errflag)
+		    breaks = loops;
+	    } else {
 		breaks = loops;
-	} else if (WTERMSIG(status) == SIGINT ||
-		   WTERMSIG(status) == SIGQUIT) {
-	    breaks = loops;
-	    errflag = 1;
+		errflag = 1;
+	    }
 	}
     }
 }
@@ -221,7 +226,7 @@ printjob(Job jn, int lng, int synch)
 {
     Process pn;
     int job = jn - jobtab, len = 9, sig, sflag = 0, llen;
-    int conted = 0, lineleng = COLUMNS, skip = 0, doputnl = 0;
+    int conted = 0, lineleng = columns, skip = 0, doputnl = 0;
     FILE *fout = (synch == 2) ? stdout : shout;
 
     if (jn->stat & STAT_NOPRINT)
@@ -249,9 +254,6 @@ printjob(Job jn, int lng, int synch)
 		    len = llen;
 		if (sig != SIGINT && sig != SIGPIPE)
 		    sflag = 1;
-		else if (sig == SIGINT && !(sigtrapped[SIGINT] & ZSIG_IGNORED))
-		    /* PWS 1995/05/16 added test for ignoring SIGINT */
-		    errflag = 1;
 		if (job == thisjob && sig == SIGINT)
 		    doputnl = 1;
 	    } else if (WIFSTOPPED(pn->status)) {
@@ -366,25 +368,34 @@ printjob(Job jn, int lng, int synch)
 	jn->stat &= ~STAT_CHANGED;
 }
 
+/**/
+void
+deletefilelist(LinkList file_list)
+{
+    char *s;
+    if (file_list) {
+	while ((s = (char *)getlinknode(file_list))) {
+	    unlink(s);
+	    zsfree(s);
+	}
+	zfree(file_list, sizeof(struct linklist));
+    }
+}
+
 /**/
 void
 deletejob(Job jn)
 {
     struct process *pn, *nx;
-    char *s;
 
     for (pn = jn->procs; pn; pn = nx) {
 	nx = pn->next;
 	zfree(pn, sizeof(struct process));
     }
     zsfree(jn->pwd);
-    if (jn->filelist) {
-	while ((s = (char *)getlinknode(jn->filelist))) {
-	    unlink(s);
-	    zsfree(s);
-	}
-	zfree(jn->filelist, sizeof(struct linklist));
-    }
+
+    deletefilelist(jn->filelist);
+
     if (jn->ty)
 	zfree(jn->ty, sizeof(struct ttyinfo));
 
@@ -629,6 +640,33 @@ spawnjob(void)
     thisjob = -1;
 }
 
+static long clktck = 0;
+
+static void
+set_clktck(void)
+{
+#ifdef _SC_CLK_TCK
+    if (!clktck)
+	/* fetch clock ticks per second from *
+	 * sysconf only the first time       */
+	clktck = sysconf(_SC_CLK_TCK);
+#else
+# ifdef __NeXT__
+    /* NeXTStep 3.3 defines CLK_TCK wrongly */
+    clktck = 60;
+# else
+#  ifdef CLK_TCK
+    clktck = CLK_TCK;
+#  else
+#   ifdef HZ
+     clktck = HZ;
+#   else
+     clktck = 60;
+#   endif
+#  endif
+# endif
+#endif
+}
 
 /* Check whether shell should report the amount of time consumed   *
  * by job.  This will be the case if we have preceded the command  *
@@ -639,7 +677,6 @@ spawnjob(void)
 int
 should_report_time(Job j)
 {
-    static long clktck = 0;
     Value v;
     char *s = "REPORTTIME";
     int reporttime;
@@ -655,23 +692,7 @@ should_report_time(Job j)
     if (!j->procs)
 	return 0;
 
-#ifdef _SC_CLK_TCK
-    if (!clktck)
-	/* fetch clock ticks per second from *
-	 * sysconf only the first time used  */
-	clktck = sysconf(_SC_CLK_TCK);
-#else
-# ifdef CLK_TCK
-    clktck = CLK_TCK;
-# else
-#  ifdef HZ
-    clktck = HZ;
-#  else
-    clktck = 60;
-#  endif
-# endif
-#endif
-
+    set_clktck();
     return ((j->procs->ti.ut + j->procs->ti.st) / clktck >= reporttime);
 }
 
@@ -699,28 +720,11 @@ printtime(struct timeval *real, struct timeinfo *ti, char *desc)
     char *s;
     double elapsed_time, user_time, system_time;
     int percent;
-    static long clktck = 0;
-
-#ifdef _SC_CLK_TCK
-    if (!clktck)
-	/* fetch clock ticks per second from *
-	 * sysconf only the first time       */
-	clktck = sysconf(_SC_CLK_TCK);
-#else
-# ifdef CLK_TCK
-    clktck = CLK_TCK;
-# else
-#  ifdef HZ
-    clktck = HZ;
-#  else
-    clktck = 60;
-#  endif
-# endif
-#endif
 
     if (!desc)
 	desc = "";
 
+    set_clktck();
     /* go ahead and compute these, since almost every TIMEFMT will have them */
     elapsed_time = real->tv_sec + real->tv_usec / 1000000.0;
     user_time    = ti->ut / (double) clktck;
