@@ -1,5 +1,5 @@
 /*
- * $Id: lex.c,v 2.42 1996/08/02 13:35:46 hzoli Exp $
+ * $Id: lex.c,v 2.47 1996/10/15 20:16:35 hzoli Exp $
  *
  * lex.c - lexical analysis
  *
@@ -10,9 +10,9 @@
  *
  * Permission is hereby granted, without written agreement and without
  * license or royalty fees, to use, copy, modify, and distribute this
- * software and its documentation for any purpose, provided that the
- * above copyright notice and the following two paragraphs appear in
- * all copies of this software.
+ * software and to distribute modified versions of this software for any
+ * purpose, provided that the above copyright notice and the following
+ * two paragraphs appear in all copies of this software.
  *
  * In no event shall Paul Falstad or the Zsh Development Group be liable
  * to any party for direct, indirect, special, incidental, or consequential
@@ -33,7 +33,6 @@
 
 /* lexical state */
 
-static int dbparens;
 int len = 0, bsiz = 256;
 char *bptr;
 extern int hwgetword;
@@ -44,11 +43,9 @@ struct lexstack {
     int incmdpos;
     int incond;
     int incasepat;
-    int dbparens;
-    int alstat;
     int isfirstln;
     int isfirstch;
-    int histremmed;
+    int histactive;
     int histdone;
     int spaceflag;
     int stophist;
@@ -88,17 +85,14 @@ lexsave(void)
 {
     struct lexstack *ls;
 
-    DPUTS(alstackind, "BUG: lexsave called with alstackind != 0");
     ls = (struct lexstack *)malloc(sizeof(struct lexstack));
 
     ls->incmdpos = incmdpos;
     ls->incond = incond;
     ls->incasepat = incasepat;
-    ls->dbparens = dbparens;
-    ls->alstat = alstat;
     ls->isfirstln = isfirstln;
     ls->isfirstch = isfirstch;
-    ls->histremmed = histremmed;
+    ls->histactive = histactive;
     ls->histdone = histdone;
     ls->spaceflag = spaceflag;
     ls->stophist = stophist;
@@ -140,11 +134,9 @@ lexrestore(void)
     incmdpos = lstack->incmdpos;
     incond = lstack->incond;
     incasepat = lstack->incasepat;
-    dbparens = lstack->dbparens;
-    alstat = lstack->alstat;
     isfirstln = lstack->isfirstln;
     isfirstch = lstack->isfirstch;
-    histremmed = lstack->histremmed;
+    histactive = lstack->histactive;
     histdone = lstack->histdone;
     spaceflag = lstack->spaceflag;
     stophist = lstack->stophist;
@@ -166,7 +158,6 @@ lexrestore(void)
     hwgetword = lstack->hwgetword;
     lexstop = lstack->lexstop;
     hdocs = lstack->hdocs;
-    clearalstack();
     hlinesz = lstack->hlinesz;
     errflag = 0;
 
@@ -315,6 +306,8 @@ initlextabs(void)
     lextok2['['] = Inbrack;
     lextok2['$'] = String;
     lextok2['~'] = Tilde;
+    lextok2['#'] = Pound;
+    lextok2['^'] = Hat;
 }
 
 /* initialize lexical state */
@@ -324,16 +317,9 @@ void
 lexinit(void)
 {
     incond = incasepat = nocorrect =
-    dbparens = alstat = lexstop = 0;
+    lexstop = 0;
     incmdpos = 1;
     tok = ENDINPUT;
-    if (isset(EXTENDEDGLOB)) {
-	lextok2['#'] = Pound;
-	lextok2['^'] = Hat;
-    } else {
-	lextok2['#'] = '#';
-	lextok2['^'] = '^';
-    }
 }
 
 /* add a char to the string buffer */
@@ -354,14 +340,63 @@ add(int c)
     }
 }
 
-#define SETPARBEGIN {if (zleparse && !alstackind && cs >= ll+1-inbufct) parbegin = inbufct;}
+#define SETPARBEGIN {if (zleparse && !(inbufflags & INP_ALIAS) && cs >= ll+1-inbufct) parbegin = inbufct;}
 #define SETPAREND {\
-	    if (zleparse && !alstackind && parbegin != -1 && parend == -1)\
+	    if (zleparse && !(inbufflags & INP_ALIAS) && parbegin != -1 && parend == -1)\
 		if (cs >= ll + 1 - inbufct)\
 		    parbegin = -1;\
 		else\
 		    parend = inbufct;}
 
+static int
+cmd_or_math(int cs_type)
+{
+    int oldlen = len;
+    int c;
+
+    cmdpush(cs_type);
+    c = dquote_parse(')', 0);
+    cmdpop();
+    *bptr = '\0';
+    if (!c) {
+	c = hgetc();
+	if (c == ')')
+	    return 1;
+	hungetc(c);
+	lexstop = 0;
+	c = ')';
+    }
+    hungetc(c);
+    lexstop = 0;
+    while (len > oldlen) {
+	len--;
+	hungetc(itok(*--bptr) ? ztokens[*bptr - Pound] : *bptr);
+    }
+    hungetc('(');
+    return 0;
+}
+
+static int
+cmd_or_math_sub(void)
+{
+    int c = hgetc();
+
+    if (c == '(') {
+	add(Inpar);
+	add('(');
+	if (cmd_or_math(CS_MATHSUBST)) {
+	    add(')');
+	    return 0;
+	}
+	bptr -= 2;
+	len -= 2;
+    } else {
+	hungetc(c);
+	lexstop = 0;
+    }
+    return skipcomm();
+}
+
 /**/
 int
 gettok(void)
@@ -377,23 +412,9 @@ gettok(void)
 	return (errflag) ? LEXERR : ENDINPUT;
     isfirstln = 0;
     wordbeg = inbufct - (qbang && c == bangchar);
-    hwbegin(-1);		/* word includes the last character read */
-    if (dbparens) {
-	len = 0;
-	bptr = tokstr = (char *)ncalloc(bsiz = 256);
-	hungetc(c);
-	cmdpush(CS_MATH);
-	c = dquote_parse(')', 0);
-	cmdpop();
-	if (c || (c = hgetc()) != ')') {
-	    hungetc(c);
-	    *bptr = '\0';
-	    return LEXERR;
-	}
-	dbparens = 0;
-	*bptr = '\0';
-	return DOUTPAR;
-    } else if (idigit(c)) {	/* handle 1< foo */
+    hwbegin(-1-(qbang && c == bangchar));
+    /* word includes the last character read and possibly \ before ! */
+    if (idigit(c)) {	/* handle 1< foo */
 	d = hgetc();
 	if (d == '>' || d == '<') {
 	    peekfd = c - '0';
@@ -484,8 +505,9 @@ gettok(void)
     case LX1_INPAR:
 	d = hgetc();
 	if (d == '(' && incmdpos) {
-	    dbparens = 1;
-	    return DINPAR;
+	    len = 0;
+	    bptr = tokstr = (char *)ncalloc(bsiz = 256);
+	    return cmd_or_math(CS_MATH) ? DINPAR : INPAR;
 	} else if (d == ')')
 	    return INOUTPAR;
 	hungetc(d);
@@ -693,23 +715,7 @@ gettokstr(int c, int sub)
 		c = Outbrack;
 	    } else if (e == '(') {
 		add(String);
-		c = hgetc();
-		if (c == '(') {
-		    add(Inpar);
-		    add('(');
-		    cmdpush(CS_MATHSUBST);
-		    if (!(c = dquote_parse(')', sub))) {
-			add(')');
-			if ((c = hgetc()) == ')')
-			    c = 0;
-		    }
-		    cmdpop();
-		}
-		else {
-		    hungetc(c);
-		    lexstop = 0;
-		    c = skipcomm();
-		}
+		c = cmd_or_math_sub();
 		if (c) {
 		    peek = LEXERR;
 		    goto brk;
@@ -787,10 +793,7 @@ gettokstr(int c, int sub)
 	    if ((isset(IGNOREBRACES) || sub) && !in_brace_param)
 		break;
 	    if (!bct)
-		if (len)
-		    goto brk;
-		else
-		    break;
+		break;
 	    if (in_brace_param)
 		cmdpop();
 	    if (bct-- == in_brace_param)
@@ -864,15 +867,28 @@ gettokstr(int c, int sub)
 		}
 	    } else if (!sub && peek != ENVSTRING &&
 		       incmdpos && !bct && !brct) {
-		e = hgetc();
-		if (e == '(' && incmdpos) {
-		    *bptr = '\0';
-		    return ENVARRAY;
+		char *t = tokstr;
+		if (idigit(*t))
+		    while (++t < bptr && idigit(*t));
+		else {
+		    while (iident(*t) && ++t < bptr);
+		    if (t < bptr) {
+			*bptr = '\0';
+			skipparens(Inbrack, Outbrack, &t);
+		    }
 		}
-		hungetc(e);
-		lexstop = 0;
-		peek = ENVSTRING;
-		intpos = 2;
+		if (t == bptr) {
+		    e = hgetc();
+		    if (e == '(' && incmdpos) {
+			*bptr = '\0';
+			return ENVARRAY;
+		    }
+		    hungetc(e);
+		    lexstop = 0;
+		    peek = ENVSTRING;
+		    intpos = 2;
+		} else
+		    c = Equals;
 	    } else
 		c = Equals;
 	    break;
@@ -976,12 +992,19 @@ gettokstr(int c, int sub)
     }
   brk:
     hungetc(c);
-    *bptr = '\0';
     if (in_brace_param) {
 	while(bct-- >= in_brace_param)
 	    cmdpop();
 	zerr("closing brace expected", NULL, 0);
+    } else if (unset(IGNOREBRACES) && !sub && len > 1 &&
+	       peek == STRING && bptr[-1] == '}' && bptr[-2] != Bnull) {
+	/* hack to get {foo} command syntax work */
+	bptr--;
+	len--;
+	lexstop = 0;
+	hungetc('}');
     }
+    *bptr = '\0';
     DPUTS(cmdsp != ocmdsp, "BUG: gettok: cmdstack changed.");
     return peek;
 }
@@ -1025,23 +1048,7 @@ dquote_parse(char endchar, int sub)
 	    c = hgetc();
 	    if (c == '(') {
 		add(Qstring);
-		c = hgetc();
-		if (c == '(') {
-		    add(Inpar);
-		    add('(');
-		    cmdpush(CS_MATHSUBST);
-		    if (!(err = dquote_parse(')', sub))) {
-			add(')');
-			if ((err = hgetc()) == ')')
-			    err = 0;
-		    }
-		    cmdpop();
-		}
-		else {
-		    hungetc(c);
-		    lexstop = 0;
-		    err = skipcomm();
-		}
+		err = cmd_or_math_sub();
 		c = Outpar;
 	    } else if (c == '[') {
 		add(String);
@@ -1144,7 +1151,7 @@ parsestr(char *s)
     HEAPALLOC {
 	lexsave();
 	untokenize(s);
-	inpush(dupstring(s), 0);
+	inpush(dupstring(s), 0, NULL);
 	strinbeg();
 	stophist = 2;
 	len = 0;
@@ -1181,7 +1188,7 @@ parse_subst_string(char *s)
 	return 0;
     lexsave();
     untokenize(s);
-    inpush(dupstring(s), 0);
+    inpush(dupstring(s), 0, NULL);
     strinbeg();
     stophist = 2;
     len = 0;
@@ -1222,7 +1229,7 @@ exalias(void)
 
     hwend();
     if (interact && isset(SHINSTDIN) && !strin && !incasepat &&
-	tok == STRING && !nocorrect && !alstackind &&
+	tok == STRING && !nocorrect && !(inbufflags & INP_ALIAS) &&
 	(isset(CORRECTALL) || (isset(CORRECT) && incmdpos)))
 	spckword(&tokstr, 1, incmdpos, 1);
 
@@ -1241,11 +1248,11 @@ exalias(void)
     } else
 	yytext = tokstr;
 
-    if (zleparse && !alstackind) {
+    if (zleparse && !(inbufflags & INP_ALIAS)) {
 	int zp = zleparse;
 
 	gotword();
-	if (zp && !zleparse) {
+	if (zp == 1 && !zleparse) {
 	    return 0;
 	}
     }
@@ -1253,12 +1260,9 @@ exalias(void)
     if (tok == STRING) {
 	/* Check for an alias */
 	an = noaliases ? NULL : (Alias) aliastab->getnode(aliastab, yytext);
-	if (alstackind != MAXAL && an && !an->inuse &&
-	    ((an->flags & ALIAS_GLOBAL) || incmdpos ||
-	     alstat == ALSTAT_MORE)) {
-	    an->inuse = 1;
-	    inpush((alstack[alstackind++] = an)->text, INP_ALIAS);
-	    alstat = 0;
+	if (an && !an->inuse && ((an->flags & ALIAS_GLOBAL) || incmdpos ||
+	     inalmore)) {
+	    inpush(an->text, INP_ALIAS, an);
 	    /* remove from history if it begins with space */
 	    if (isset(HISTIGNORESPACE) && an->text[0] == ' ')
 		remhist();
@@ -1267,7 +1271,8 @@ exalias(void)
 	}
 
 	/* Then check for a reserved word */
-	if ((incmdpos || (yytext[0] == '}' && !yytext[1])) &&
+	if ((incmdpos ||
+	     (unset(IGNOREBRACES) && yytext[0] == '}' && !yytext[1])) &&
 	    (rw = (Reswd) reswdtab->getnode(reswdtab, yytext))) {
 	    tok = rw->token;
 	    if (tok == DINBRACK)
@@ -1278,6 +1283,7 @@ exalias(void)
 	} else if (incond && yytext[0] == '!' && !yytext[1])
 	    tok = BANG;
     }
+    inalmore = 0;
     return 0;
 }
 
