@@ -1,5 +1,5 @@
 /*
- * $Id: mem.c,v 2.10 1996/08/04 12:35:44 hzoli Exp $
+ * $Id: mem.c,v 2.13 1996/10/15 20:16:35 hzoli Exp $
  *
  * mem.c - memory management
  *
@@ -10,9 +10,9 @@
  *
  * Permission is hereby granted, without written agreement and without
  * license or royalty fees, to use, copy, modify, and distribute this
- * software and its documentation for any purpose, provided that the
- * above copyright notice and the following two paragraphs appear in
- * all copies of this software.
+ * software and to distribute modified versions of this software for any
+ * purpose, provided that the above copyright notice and the following
+ * two paragraphs appear in all copies of this software.
  *
  * In no event shall Paul Falstad or the Zsh Development Group be liable
  * to any party for direct, indirect, special, incidental, or consequential
@@ -92,6 +92,7 @@ int h_m[1025], h_push, h_pop, h_free;
 
 #define H_ISIZE  sizeof(long)
 #define HEAPSIZE (8192 - H_ISIZE)
+#define HEAP_ARENA_SIZE (HEAPSIZE - sizeof(struct heap))
 #define HEAPFREE (16384 - H_ISIZE)
 
 /* set default allocation to heap stack */
@@ -122,24 +123,20 @@ global_permalloc(void)
     return luh;
 }
 
-/* Information about the state of a heap.
-   Each heap keeps a list of these structures.  Pushing a new heap just
-   means to save the current state of each heap in one of these structures.
-   Popping a heap means to restore the last state of each heap and freeing
-   the structure that held the state. */
+/* heappush saves the current heap state using this structure */
 
-struct heappos {
-    struct heappos *next;	/* next one in list for this heap */
-    char *ptr;			/* pointer to free area on heap   */
-    size_t free;       		/* number of free bytes on heap   */
+struct heapstack {
+    struct heapstack *next;	/* next one in list for this heap */
+    size_t used;
 };
 
 /* A zsh heap. */
 
 struct heap {
-    struct heap *next;		/* next one                               */
-    struct heappos pos;		/* list of structures keeping heap states */
-    char *arena;		/* the memory area of this heap           */
+    struct heap *next;		/* next one                                  */
+    size_t used;		/* bytes used from the heap                  */
+    struct heapstack *sp;	/* used by pushheap() to save the value used */
+#define arena(X)	((char *) (X) + sizeof(struct heap))
 };
 
 /* list of zsh heap */
@@ -153,18 +150,17 @@ void
 pushheap(void)
 {
     Heap h;
-    Heappos hp;
+    Heapstack hs;
 
 #if defined(ZSH_MEM) && defined(ZSH_MEM_DEBUG)
     h_push++;
 #endif
 
     for (h = heaps; h; h = h->next) {
-	hp = (Heappos) zalloc(sizeof(*hp));
-	hp->next = h->pos.next;
-	h->pos.next = hp;
-	hp->free = h->pos.free;
-	hp->ptr = h->pos.ptr;
+	hs = (Heapstack) zalloc(sizeof(*hs));
+	hs->next = h->sp;
+	h->sp = hs;
+	hs->used = h->used;
     }
 }
 
@@ -180,13 +176,13 @@ freeheap(void)
     h_free++;
 #endif
     for (h = heaps; h; h = h->next) {
-	if (h->pos.next) {
-	    h->pos.free = h->pos.next->free;
-	    h->pos.ptr = h->pos.next->ptr;
-	} else {
-	    h->pos.free += (int)(h->pos.ptr - h->arena);
-	    h->pos.ptr = h->arena;
-	}
+#ifdef ZSH_MEM_DEBUG
+	if (h->sp)
+	    memset(arena(h) + h->sp->used, 0xff, h->used - h->sp->used);
+	else
+	    memset(arena(h), 0xff, h->used);
+#endif
+	h->used = h->sp ? h->sp->used : 0;
     }
 }
 
@@ -197,7 +193,7 @@ void
 popheap(void)
 {
     Heap h, hn, hl = NULL;
-    Heappos hp, hpn;
+    Heapstack hs;
 
 #if defined(ZSH_MEM) && defined(ZSH_MEM_DEBUG)
     h_pop++;
@@ -205,21 +201,17 @@ popheap(void)
 
     for (h = heaps; h; h = hn) {
 	hn = h->next;
-	if ((hp = h->pos.next)) {
-	    h->pos.next = hp->next;
-	    h->pos.free = hp->free;
-	    h->pos.ptr = hp->ptr;
-	    zfree(hp, sizeof(struct heappos));
+	if ((hs = h->sp)) {
+	    h->sp = hs->next;
+#ifdef ZSH_MEM_DEBUG
+	    memset(arena(h) + hs->used, 0xff, h->used - hs->used);
+#endif
+	    h->used = hs->used;
+	    zfree(hs, sizeof(*hs));
 
 	    hl = h;
-	} else {
-	    for (hp = h->pos.next; hp; hp = hpn) {
-		hpn = hp->next;
-		free(hp);
-	    }
-	    zfree(h->arena, HEAPSIZE);
-	    zfree(h, sizeof(struct heap));
-	}
+	} else
+	    zfree(h, HEAPSIZE);
     }
     if (hl)
 	hl->next = NULL;
@@ -233,10 +225,8 @@ popheap(void)
 void *
 halloc(size_t size)
 {
-    Heap h, hp;
-    char *ret;
-
-    queue_signals();  /* just queue signals caught rather than handling them */
+    Heap h;
+    size_t n;
 
     size = (size + H_ISIZE - 1) & ~(H_ISIZE - 1);
 
@@ -246,15 +236,15 @@ halloc(size_t size)
 
     /* find a heap with enough free space */
 
-    for (h = heaps; h && h->pos.free < size; h = h->next);
+    for (h = heaps; h; h = h->next) {
+	if (HEAP_ARENA_SIZE >= (n = size + h->used)) {
+	    h->used = n;
+	    return arena(h) + n - size;
+	}
+    }
 
-    if (h) {
-        /* found */
-	ret = h->pos.ptr;
-	h->pos.ptr += size;
-	h->pos.free -= size;
-    } else {
-	size_t n = (size > HEAPSIZE) ? size : HEAPSIZE;
+    {
+	Heap hp;
         /* not found, allocate new heap */
 #ifdef ZSH_MEM
 	static int called = 0;
@@ -262,10 +252,11 @@ halloc(size_t size)
             /* tricky, see above */
 #endif
 
+	queue_signals();
+	n = HEAP_ARENA_SIZE > size ? HEAP_ARENA_SIZE : size;
 	for (hp = NULL, h = heaps; h; hp = h, h = h->next);
 
-	h = (Heap) zalloc(sizeof *h);
-	h->arena = (char *)zalloc(n);
+	h = (Heap) zalloc(n + sizeof(*h));
 
 #ifdef ZSH_MEM
 	if (called)
@@ -273,23 +264,18 @@ halloc(size_t size)
 	called = 1;
 #endif
 
+	h->used = size;
 	h->next = NULL;
+	h->sp = NULL;
 
 	if (hp)
 	    hp->next = h;
 	else
 	    heaps = h;
 
-	h->pos.next = NULL;
-	h->pos.ptr = h->arena + size;
-
-	h->pos.free = n - size;
-
-	ret = h->arena;
+	unqueue_signals();
+	return arena(h);
     }
-
-    unqueue_signals();       /* stop queueing signals and process any that were queued */
-    return (void *) ret;
 }
 
 /* allocate memory from the current memory pool and clear it */
@@ -631,19 +617,19 @@ malloc(MALLOC_ARG_T size)
     that the blocks on the free list are ordered) */
 
     for (hp = NULL, h = heaps; h; hp = h, h = h->next)
-	if (h->pos.ptr == h->arena &&
+	if (!h->used &&
 	    (!hf || h < hf) &&
 	    (!m || ((char *)m) > ((char *)h)))
 	    hf = h, hfp = hp;
 
     if (hf) {
 	/* we found such a heap */
-	struct heappos *hpo, *hpn;
+	Heapstack hso, hsn;
 
 	/* delete structures on the list holding the heap states */
-	for (hpo = hf->pos.next; hpo; hpo = hpn) {
-	    hpn = hpo->next;
-	    zfree(hpo, sizeof(struct heappos));
+	for (hso = hf->sp; hso; hso = hsn) {
+	    hsn = hso->next;
+	    zfree(hso, sizeof(*hso));
 	}
 	/* take it from the list of heaps */
 	if (hfp)
@@ -651,8 +637,7 @@ malloc(MALLOC_ARG_T size)
 	else
 	    heaps = hf->next;
 	/* now we simply free it and than search the free list again */
-	zfree(hf->arena, HEAPSIZE);
-	zfree(hf, sizeof(struct heap));
+	zfree(hf, HEAPSIZE);
 
 	for (mp = NULL, m = m_free; m && m->len < size; mp = m, m = m->next);
     }
