@@ -1,5 +1,5 @@
 /*
- * $Id: params.c,v 2.34 1996/08/04 12:52:33 hzoli Exp $
+ * $Id: params.c,v 2.41 1996/10/15 20:16:35 hzoli Exp $
  *
  * params.c - parameters
  *
@@ -10,9 +10,9 @@
  *
  * Permission is hereby granted, without written agreement and without
  * license or royalty fees, to use, copy, modify, and distribute this
- * software and its documentation for any purpose, provided that the
- * above copyright notice and the following two paragraphs appear in
- * all copies of this software.
+ * software and to distribute modified versions of this software for any
+ * purpose, provided that the above copyright notice and the following
+ * two paragraphs appear in all copies of this software.
  *
  * In no event shall Paul Falstad or the Zsh Development Group be liable
  * to any party for direct, indirect, special, incidental, or consequential
@@ -41,8 +41,7 @@ static Param argvparam;
 void
 createparamtable(void)
 {
-    struct iparam *ip;
-    Param pm;
+    Param ip, pm;
     char **old_environ, **envp, **envp2, **sigptr, **t;
     char buf[50], *str, *iname;
     int num_env;
@@ -76,86 +75,88 @@ createparamtable(void)
 
     noerrs = 1;
 
-    /* Add the standard non-special parameters which have to    *
-     * be initialized before we copy the environment variables. *
-     * We don't want to override whatever values the users has  *
-     * given them in the environment.                           */
-    setiparam("MAILCHECK", 60);
-    setiparam("LOGCHECK", 60);
-    setiparam("KEYTIMEOUT", 40);
-    setiparam("LISTMAX", 100);
+    HEAPALLOC {
+	/* Add the standard non-special parameters which have to    *
+	 * be initialized before we copy the environment variables. *
+	 * We don't want to override whatever values the users has  *
+	 * given them in the environment.                           */
+	setiparam("MAILCHECK", 60);
+	setiparam("LOGCHECK", 60);
+	setiparam("KEYTIMEOUT", 40);
+	setiparam("LISTMAX", 100);
 #ifdef HAVE_SELECT
-    setiparam("BAUD", getbaudrate(&shttyinfo));  /* get the output baudrate */
+	setiparam("BAUD", getbaudrate(&shttyinfo));  /* get the output baudrate */
 #endif
-    setsparam("FCEDIT", ztrdup(DEFAULT_FCEDIT));
-    setsparam("TMPPREFIX", ztrdup(DEFAULT_TMPPREFIX));
-    setsparam("TIMEFMT", ztrdup(DEFAULT_TIMEFMT));
-    setsparam("WATCHFMT", ztrdup(DEFAULT_WATCHFMT));
-    setsparam("HOST", ztrdup(hostnam));
-    setsparam("LOGNAME", ztrdup((str = getlogin()) && *str ? str : cached_username));
+	setsparam("FCEDIT", ztrdup(DEFAULT_FCEDIT));
+	setsparam("TMPPREFIX", ztrdup(DEFAULT_TMPPREFIX));
+	setsparam("TIMEFMT", ztrdup(DEFAULT_TIMEFMT));
+	setsparam("WATCHFMT", ztrdup(DEFAULT_WATCHFMT));
+	setsparam("HOST", ztrdup(hostnam));
+	setsparam("LOGNAME", ztrdup((str = getlogin()) && *str ? str : cached_username));
 
-    /* Copy the environment variables we are inheriting to dynamic *
-     * memory, so we can do mallocs and frees on it.               */
-    old_environ = environ;
-    num_env = arrlen(old_environ);
-    environ = (char **) zalloc(sizeof(char *) * (num_env + 1));
-    for (envp = environ, envp2 = old_environ; *envp2; envp2++)
-	*envp++ = ztrdup(*envp2);
-    *envp = NULL;
+	/* Copy the environment variables we are inheriting to dynamic *
+	 * memory, so we can do mallocs and frees on it.               */
+	old_environ = environ;
+	num_env = arrlen(old_environ);
+	environ = (char **) zalloc(sizeof(char *) * (num_env + 1));
+	for (envp = environ, envp2 = old_environ; *envp2; envp2++)
+	    *envp++ = ztrdup(*envp2);
+	*envp = NULL;
 
-    /* Now incorporate environment variables we are inheriting *
-     * into the parameter hash table.                          */
-    for (envp = environ, envp2 = old_environ; *envp2; envp++, envp2++) {
-	for (str = *envp2; *str && *str != '='; str++);
-	if (*str == '=') {
-	    iname = NULL;
-	    *str = '\0';
-	    if (!idigit(**envp2) && isident(*envp2) && !strchr(*envp2, '[')) {
-		iname = *envp2;
-		if ((!(pm = (Param) paramtab->getnode(paramtab, iname)) ||
-		     !(pm->flags & PM_DONTIMPORT)) &&
-		    (pm = setsparam(iname, metafy(str + 1, -1, META_DUP)))) {
-		    pm->flags |= PM_EXPORTED;
-		    pm->env = *envp;
-		    if (pm->flags & PM_SPECIAL)
-			pm->env = replenv(pm->env, getsparam(iname));
-		} else if (pm)
-		    for (t = envp--; (t[0] = t[1]); t++);
+	/* Now incorporate environment variables we are inheriting *
+	 * into the parameter hash table.                          */
+	for (envp = environ, envp2 = old_environ; *envp2; envp++, envp2++) {
+	    for (str = *envp2; *str && *str != '='; str++);
+	    if (*str == '=') {
+		iname = NULL;
+		*str = '\0';
+		if (!idigit(**envp2) && isident(*envp2) && !strchr(*envp2, '[')) {
+		    iname = *envp2;
+		    if ((!(pm = (Param) paramtab->getnode(paramtab, iname)) ||
+			 !(pm->flags & PM_DONTIMPORT)) &&
+			(pm = setsparam(iname, metafy(str + 1, -1, META_DUP)))) {
+			pm->flags |= PM_EXPORTED;
+			pm->env = *envp;
+			if (pm->flags & PM_SPECIAL)
+			    pm->env = replenv(pm->env, getsparam(iname));
+		    } else if (pm)
+			for (t = envp--; (t[0] = t[1]); t++);
+		}
+		*str = '=';
 	    }
-	    *str = '=';
 	}
-    }
 
-    pm = (Param) paramtab->getnode(paramtab, "HOME");
-    if (!(pm->flags & PM_EXPORTED)) {
-	pm->flags |= PM_EXPORTED;
-	pm->env = addenv("HOME", home);
-    }
-    pm = (Param) paramtab->getnode(paramtab, "PWD");
-    if (!(pm->flags & PM_EXPORTED)) {
-	pm->flags |= PM_EXPORTED;
-	pm->env = addenv("PWD", pwd);
-    }
-    pm = (Param) paramtab->getnode(paramtab, "LOGNAME");
-    if (!(pm->flags & PM_EXPORTED)) {
-	pm->flags |= PM_EXPORTED;
-	pm->env = addenv("LOGNAME", pm->u.str);
-    }
-    pm = (Param) paramtab->getnode(paramtab, "SHLVL");
-    if (!(pm->flags & PM_EXPORTED))
-	pm->flags |= PM_EXPORTED;
-    sprintf(buf, "%d", (int)++shlvl);
-    pm->env = addenv("SHLVL", buf);
+	pm = (Param) paramtab->getnode(paramtab, "HOME");
+	if (!(pm->flags & PM_EXPORTED)) {
+	    pm->flags |= PM_EXPORTED;
+	    pm->env = addenv("HOME", home);
+	}
+	pm = (Param) paramtab->getnode(paramtab, "PWD");
+	if (!(pm->flags & PM_EXPORTED)) {
+	    pm->flags |= PM_EXPORTED;
+	    pm->env = addenv("PWD", pwd);
+	}
+	pm = (Param) paramtab->getnode(paramtab, "LOGNAME");
+	if (!(pm->flags & PM_EXPORTED)) {
+	    pm->flags |= PM_EXPORTED;
+	    pm->env = addenv("LOGNAME", pm->u.str);
+	}
+	pm = (Param) paramtab->getnode(paramtab, "SHLVL");
+	if (!(pm->flags & PM_EXPORTED))
+	    pm->flags |= PM_EXPORTED;
+	sprintf(buf, "%d", (int)++shlvl);
+	pm->env = addenv("SHLVL", buf);
 
-    /* Add the standard non-special parameters */
-    setsparam("MACHTYPE", ztrdup(MACHTYPE));
-    setsparam("OSTYPE", ztrdup(OSTYPE));
-    setsparam("TTY", ztrdup(ttystrname));
-    setsparam("VENDOR", ztrdup(VENDOR));
-    setsparam("ZSH_NAME", ztrdup(zsh_name));
-    setsparam("ZSH_VERSION", ztrdup(ZSH_VERSION));
-    setaparam("signals", sigptr = zalloc((SIGCOUNT+4) * sizeof(char *)));
-    for (t = sigs; (*sigptr++ = ztrdup(*t++)); );
+	/* Add the standard non-special parameters */
+	setsparam("MACHTYPE", ztrdup(MACHTYPE));
+	setsparam("OSTYPE", ztrdup(OSTYPE));
+	setsparam("TTY", ztrdup(ttystrname));
+	setsparam("VENDOR", ztrdup(VENDOR));
+	setsparam("ZSH_NAME", ztrdup(zsh_name));
+	setsparam("ZSH_VERSION", ztrdup(ZSH_VERSION));
+	setaparam("signals", sigptr = zalloc((SIGCOUNT+4) * sizeof(char *)));
+	for (t = sigs; (*sigptr++ = ztrdup(*t++)); );
+    } LASTALLOC;
 
     noerrs = 0;
 }
@@ -174,15 +175,20 @@ createparam(char *name, int flags)
 	spec = oldpm && (oldpm->flags & PM_SPECIAL);
 
 	if ((oldpm && oldpm->level == locallevel) || spec) {
+	    if (oldpm && !(oldpm->flags & PM_UNSET))
+		return NULL;
 	    pm = oldpm;
 	    pm->ct = 0;
 	    oldpm = pm->old;
-	    pm->flags = (flags & (PM_EXPORTED | PM_LEFT | PM_RIGHT_B | PM_RIGHT_Z |
-				  PM_LOWER | PM_UPPER | PM_READONLY | PM_TAGGED)) |
+	    pm->flags = (flags & (PM_EXPORTED | PM_LEFT | PM_RIGHT_B |
+				  PM_RIGHT_Z | PM_LOWER | PM_UPPER |
+				  PM_READONLY | PM_TAGGED | PM_UNIQUE)) |
 		(pm->flags & (PM_SCALAR | PM_INTEGER | PM_ARRAY | PM_SPECIAL));
-	    if (pm->ename && (altpm = (Param) paramtab->getnode(paramtab, pm->ename))) {
-		altpm->flags &= ~(PM_UNSET | PM_EXPORTED | PM_LEFT | PM_RIGHT_B |
-				  PM_RIGHT_Z | PM_LOWER | PM_UPPER | PM_READONLY | PM_TAGGED);
+	    if (pm->ename &&
+		(altpm = (Param) paramtab->getnode(paramtab, pm->ename))) {
+		altpm->flags &= ~(PM_UNSET | PM_UNIQUE | PM_UPPER | PM_LEFT |
+				  PM_RIGHT_B | PM_RIGHT_Z | PM_LOWER |
+				  PM_READONLY | PM_TAGGED | PM_EXPORTED);
 	    }
 	} else {
 	    pm = (Param) zcalloc(sizeof *pm);
@@ -359,10 +365,7 @@ getarg(char **str, int *inv, Value v, int a2, long *w)
 
     if (!*t)
 	return 0;
-    sav = *t;
-    *t = '\0';
-    s = dupstring(s);
-    *t = sav;
+    s = dupstrpfx(s, t - s);
     *str = tt = t;
     if (parsestr(s))
 	return 0;
@@ -404,7 +407,6 @@ getarg(char **str, int *inv, Value v, int a2, long *w)
 		r -= strlen(s);
 	    }
 	}
-		
     } else {
 	if (!v->isarr && !word) {
 	    l = strlen(s);
@@ -543,8 +545,6 @@ getindex(char **pptr, Value v)
 	    if (a > 0 && isset(KSHARRAYS))
 		a--;
 	    v->inv = 1;
-	    if (*s == ']' || *s == Outbrack)
-		s++;
 	    v->isarr = 0;
 	    v->a = v->b = a;
 	    if (*s == ',') {
@@ -554,6 +554,8 @@ getindex(char **pptr, Value v)
 		*pptr = s;
 		return 1;
 	    }
+	    if (*s == ']' || *s == Outbrack)
+		s++;
 	} else {
 	    if (a > 0)
 		a--;
@@ -675,46 +677,49 @@ getstrvalue(Value v)
     static char buf[(SIZEOF_LONG * 8) + 4];
 
     if (!v)
-	return "";
-    if (v->inv) {
-	sprintf(buf, "%d", v->a);
-	return dupstring(buf);
-    }
+	return hcalloc(1);
+    HEAPALLOC {
+	if (v->inv) {
+	    sprintf(buf, "%d", v->a);
+	    s = dupstring(buf);
+	    LASTALLOC_RETURN s;
+	}
 
-    switch(PM_TYPE(v->pm->flags)) {
-    case PM_ARRAY:
-	if (v->isarr)
-	    return sepjoin(v->pm->gets.afn(v->pm), NULL);
+	switch(PM_TYPE(v->pm->flags)) {
+	case PM_ARRAY:
+	    if (v->isarr)
+		s = sepjoin(v->pm->gets.afn(v->pm), NULL);
+	    else {
+		ss = v->pm->gets.afn(v->pm);
+		if (v->a < 0)
+		    v->a += arrlen(ss);
+		s = (v->a >= arrlen(ss) || v->a < 0) ? (char *) hcalloc(1) : ss[v->a];
+	    }
+	    LASTALLOC_RETURN s;
+	case PM_INTEGER:
+	    convbase(s = buf, v->pm->gets.ifn(v->pm), v->pm->ct);
+	    break;
+	case PM_SCALAR:
+	    s = v->pm->gets.cfn(v->pm);
+	    break;
+	default:
+	    s = NULL;
+	    DPUTS(1, "oops, param node without valid type");
+	    break;
+	}
 
-	ss = v->pm->gets.afn(v->pm);
+	if (v->a == 0 && v->b == -1)
+	    LASTALLOC_RETURN s;
 	if (v->a < 0)
-	    v->a += arrlen(ss);
-	s = (v->a >= arrlen(ss) || v->a < 0) ? "" : ss[v->a];
-	return s;
-	break;
-    case PM_INTEGER:
-	convbase(s = buf, v->pm->gets.ifn(v->pm), v->pm->ct);
-	break;
-    case PM_SCALAR:
-	s = v->pm->gets.cfn(v->pm);
-	break;
-    default:
-	s = NULL;
-	zerr("oops, param node without valid type", NULL, 0);
-	break;
-    }
-
-    if (v->a == 0 && v->b == -1)
-	return s;
-    if (v->a < 0)
-	v->a += strlen(s);
-    if (v->b < 0)
-	v->b += strlen(s);
-    s = (v->a > (int)strlen(s)) ? dupstring("") : dupstring(s + v->a);
-    if (v->b < v->a)
-	s[0] = '\0';
-    else if (v->b - v->a < (int)strlen(s))
-	s[v->b - v->a + 1] = '\0';
+	    v->a += strlen(s);
+	if (v->b < 0)
+	    v->b += strlen(s);
+	s = (v->a > (int)strlen(s)) ? dupstring("") : dupstring(s + v->a);
+	if (v->b < v->a)
+	    s[0] = '\0';
+	else if (v->b - v->a < (int)strlen(s))
+	    s[v->b - v->a + 1 + (s[v->b - v->a] == Meta)] = '\0';
+    } LASTALLOC;
     return s;
 }
 
@@ -774,11 +779,13 @@ setstrvalue(Value v, char *val)
     char buf[(SIZEOF_LONG * 8) + 4];
 
     if (v->pm->flags & PM_READONLY) {
+	zerr("read-only variable: %s", v->pm->nam, 0);
 	zsfree(val);
 	return;
     }
     switch (PM_TYPE(v->pm->flags)) {
     case PM_SCALAR:
+	MUSTUSEHEAP("setstrvalue");
 	if (v->a == 0 && v->b == -1) {
 	    (v->pm->sets.cfn) (v->pm, val);
 	    if (v->pm->flags & (PM_LEFT | PM_RIGHT_B | PM_RIGHT_Z) && !v->pm->ct)
@@ -819,6 +826,7 @@ setstrvalue(Value v, char *val)
 	    v->pm->ct = lastbase;
 	break;
     case PM_ARRAY:
+	MUSTUSEHEAP("setstrvalue");
 	{
 	    char **ss = (char **) zalloc(2 * sizeof(char *));
 
@@ -850,8 +858,10 @@ setintvalue(Value v, long val)
 {
     char buf[DIGBUFSIZE];
 
-    if (v->pm->flags & PM_READONLY)
+    if (v->pm->flags & PM_READONLY) {
+	zerr("read-only variable: %s", v->pm->nam, 0);
 	return;
+    }
     switch (PM_TYPE(v->pm->flags)) {
     case PM_SCALAR:
     case PM_ARRAY:
@@ -865,24 +875,12 @@ setintvalue(Value v, long val)
     }
 }
 
-/**/
-void
-setintenv(char *s, long val)
-{
-    Param pm;
-    char buf[DIGBUFSIZE];
-
-    if ((pm = (Param) paramtab->getnode(paramtab, s)) && pm->env) {
-	sprintf(buf, "%ld", val);
-	pm->env = replenv(pm->env, buf);
-    }
-}
-
 /**/
 void
 setarrvalue(Value v, char **val)
 {
     if (v->pm->flags & PM_READONLY) {
+	zerr("read-only variable: %s", v->pm->nam, 0);
 	freearray(val);
 	return;
     }
@@ -964,9 +962,10 @@ getaparam(char *s)
 {
     Value v;
 
-    if (!((v = getvalue(&s, 0)) && v->isarr))
-	return NULL;
-    return getarrvalue(v);
+    if (!idigit(*s) && (v = getvalue(&s, 0)) &&
+	PM_TYPE(v->pm->flags) == PM_ARRAY)
+	return v->pm->gets.afn(v->pm);
+    return NULL;
 }
 
 /**/
@@ -999,9 +998,10 @@ setsparam(char *s, char *val)
 	    v = NULL;
 	}
     }
-    if (!v)
-	if (!(v = getvalue(&t, 1)))
-	    return NULL;
+    if (!v && !(v = getvalue(&t, 1))) {
+	zsfree(val);
+	return NULL;
+    }
     setstrvalue(v, val);
     return v->pm;
 }
@@ -1061,6 +1061,7 @@ setiparam(char *s, long val)
     }
     if (!(v = getvalue(&s, 1))) {
 	pm = createparam(t, PM_INTEGER);
+	DPUTS(!pm, "BUG: parameter not created");
 	pm->u.val = val;
 	return pm;
     }
@@ -1089,8 +1090,10 @@ unsetparam_pm(Param pm, int altflag)
     Param oldpm, altpm;
     int spec;
 
-    if ((pm->flags & PM_READONLY) && pm->level <= locallevel)
+    if ((pm->flags & PM_READONLY) && pm->level <= locallevel) {
+	zerr("read-only variable: %s", pm->nam, 0);
 	return;
+    }
     spec = (pm->flags & PM_SPECIAL) && !pm->level;
     switch (PM_TYPE(pm->flags)) {
     case PM_SCALAR:
@@ -1153,7 +1156,7 @@ intsetfn(Param pm, long x)
 char *
 strgetfn(Param pm)
 {
-    return pm->u.str ? pm->u.str : "";
+    return pm->u.str ? pm->u.str : (char *) hcalloc(1);
 }
 
 /* Function to set value of a scalar (string) parameter */
@@ -1223,6 +1226,33 @@ intvarsetfn(Param pm, long x)
     *((long *)pm->data) = x;
 }
 
+/* Function to set value of any ZLE-related integer *
+ * parameter.  data is pointer to global variable   *
+ * where the value is to be stored.                 */
+
+/**/
+void
+zlevarsetfn(Param pm, long x)
+{
+    if ((long *)pm->data == & columns) {
+	if(x <= 0)
+	    x = tccolumns > 0 ? tccolumns : 80;
+	if (x > 2)
+	    termflags &= ~TERM_NARROW;
+	else
+	    termflags |= TERM_NARROW;
+    } else if ((long *)pm->data == & lines) {
+	if(x <= 0)
+	    x = tclines > 0 ? tclines : 24;
+	if (x > 2)
+	    termflags &= ~TERM_SHORT;
+	else
+	    termflags |= TERM_SHORT;
+    }
+
+    *((long *)pm->data) = x;
+}
+
 /* Function to set value of generic special scalar    *
  * parameter.  data is pointer to a character pointer *
  * representing the scalar (string).                  */
@@ -1248,7 +1278,7 @@ strvargetfn(Param pm)
     char *s = *((char **)pm->data);
 
     if (!s)
-	return "";
+	return hcalloc(1);
     return s;
 }
 
@@ -1285,24 +1315,6 @@ arrvarsetfn(Param pm, char **x)
 	arrfixenv(pm->ename, x);
 }
 
-/**/
-char **
-colonfix(char *x, char *ename, int uniq)
-{
-    char **s, **t;
-
-    s = colonsplit(x, uniq);
-    if (ename)
-	arrfixenv(ename, s);
-    zsfree(x);
-    for (t = s; *t; t++)
-	if (!**t) {
-	    free(*t);
-	    *t = ztrdup(".");
-	}
-    return s;
-}
-
 /**/
 char *
 colonarrgetfn(Param pm)
@@ -1317,29 +1329,11 @@ colonarrsetfn(Param pm, char *x)
     char ***dptr = (char ***)pm->data;
 
     freearray(*dptr);
-    if (pm->data == (void *) & mailpath && x && !*x) {
-	zsfree(x);
-	*dptr = mkarray(NULL);
-    } else
-	*dptr = x ? colonfix(x, pm->ename ? pm->nam : NULL, pm->flags & PM_UNIQUE) : mkarray(NULL);
+    *dptr = x ? colonsplit(x, pm->flags & PM_UNIQUE) : mkarray(NULL);
+    if (pm->ename)
+	arrfixenv(pm->nam, *dptr);
 }
 
-/* Function to get the value of special (scalar)        *
- * parameter `PATH'.  This parameter is not really      *
- * stored, so it gets the value by joining the elements *
- * of its associated (array) parameter `path'.          */
-
-/**/
-char *
-colonpathgetfn(Param pm)
-{
-    return zjoin(path, ':');
-}
-
-/* Function to set special (scalar) parameter `PATH'.     *
- * This actually updates its associated (array) parameter *
- * `path' which is the only one that is stored.           */
-
 /**/
 int
 uniqarray(char **x)
@@ -1360,40 +1354,6 @@ uniqarray(char **x)
     return changes;
 }
 
-/**/
-void
-colonpathsetfn(Param pm, char *x)
-{
-    freearray(path);
-    path = x ? colonfix(x, pm->nam, pm->flags & PM_UNIQUE) : mkarray(NULL);
-    cmdnamtab->emptytable(cmdnamtab);
-}
-
-/* Function to get value of special parameter `path' */
-
-/**/
-char **
-pathgetfn(Param pm)
-{
-    return path;
-}
-
-/* Function to set value of special parameter `path' */
-
-/**/
-void
-pathsetfn(Param pm, char **x)
-{
-    if (path != x)
-	freearray(path);
-    if (pm->flags & PM_UNIQUE)
-	uniqarray(x);
-    path = x ? x : mkarray(NULL);
-    cmdnamtab->emptytable(cmdnamtab);
-    if (x)
-	arrfixenv("PATH", x);
-}
-
 /* Function to get value of special parameter `#' and `ARGC' */
 
 /**/
@@ -1459,6 +1419,9 @@ usernamesetfn(Param pm, char *x)
     struct passwd *pswd;
 
     if (x && (pswd = getpwnam(x)) && (pswd->pw_uid != cached_uid)) {
+#ifdef HAVE_INITGROUPS
+	initgroups(x, pswd->pw_gid);
+#endif
 	if(!setgid(pswd->pw_gid) && !setuid(pswd->pw_uid)) {
 	    zsfree(cached_username);
 	    cached_username = ztrdup(pswd->pw_name);
@@ -1798,7 +1761,7 @@ termsetfn(Param pm, char *x)
 
     /* If non-interactive, delay setting up term till we need it. */
     if (unset(INTERACTIVE) || !*term)
-	termok = TERM_BAD;
+	termflags |= TERM_UNKNOWN;
     else 
 	init_term();
 }
@@ -1814,6 +1777,9 @@ arrfixenv(char *s, char **t)
     int len_s;
     Param pm;
 
+    MUSTUSEHEAP("arrfixenv");
+    if (t == path)
+	cmdnamtab->emptytable(cmdnamtab);
     u = zjoin(t, ':');
     len_s = strlen(s);
     pm = (Param) paramtab->getnode(paramtab, s);
