@@ -367,9 +367,9 @@ par_cmd(void)
 	cmdpop();
 	break;
     case DINPAR:
-	cmdpush(CS_MATH);
-	par_dinpar(c);
-	cmdpop();
+	c->type = CARITH;
+	addlinknode(c->args, tokstr);
+	yylex();
 	break;
     default:
 	if (!par_simple(c))
@@ -462,7 +462,7 @@ par_case(Cmd c)
 {
     int brflag;
     LinkList pats, lists;
-    int n = 0;
+    int n = 1;
     char **pp;
     List *ll;
     LinkNode no;
@@ -473,7 +473,8 @@ par_case(Cmd c)
     yylex();
     if (tok != STRING)
 	YYERRORV;
-    addlinknode(c->args, tokstr);
+    pats = newlinklist();
+    addlinknode(pats, tokstr);
     incmdpos = 1;
     yylex();
     while (tok == SEPER)
@@ -485,7 +486,6 @@ par_case(Cmd c)
     incmdpos = 0;
     yylex();
     cc = c->u.casecmd = (struct casecmd *)make_casecmd();
-    pats = newlinklist();
     lists = newlinklist();
     for (;;) {
 	char *str;
@@ -884,24 +884,6 @@ par_dinbrack(Cmd c)
     yylex();
 }
 
-/*
- * dinpar : DINPAR expr DOUTPAR
- */
-
-/**/
-void
-par_dinpar(Cmd c)
-{
-    c->type = CARITH;
-    incmdpos = 0;
-    yylex();
-    if (tok != DOUTPAR)
-	YYERRORV;
-    addlinknode(c->args, tokstr);
-    incmdpos = 1;
-    yylex();
-}
-
 /*
  * simple	: { COMMAND | EXEC | NOGLOB | NOCORRECT | DASH }
 					{ STRING | ENVSTRING | ENVARRAY wordlist OUTPAR | redir }
