@@ -422,13 +422,9 @@ handler(int sig)
 	    for (;;) {
 		if (pid == *procsubpid) {
 		    *procsubpid = 0;
-		    if (WIFSIGNALED(status)) {
+		    if (WIFSIGNALED(status))
 			*procsubval = (0200 | WTERMSIG(status));
-			if (WTERMSIG(status) == SIGINT)
-			    kill(getpid(), SIGINT);
-			else if (sigtrapped[WTERMSIG(status)])
-			    dotrap(WTERMSIG(status));
-		    } else
+		    else
 			*procsubval = WEXITSTATUS(status);
 		    times(&shtms);
 		    goto cont;
@@ -482,13 +478,18 @@ handler(int sig)
             if (list_pipe || chline || simple_pline) {
                 breaks = loops;
                 errflag = 1;
+		inerrflush();
             }
         }
         break;
 
+#ifdef SIGWINCH
     case SIGWINCH:
         adjustwinsize();  /* check window size and adjust */
+	if (sigtrapped[SIGWINCH])
+	    dotrap(SIGWINCH);
         break;
+#endif
 
     case SIGALRM:
         if (sigtrapped[SIGALRM]) {
@@ -588,8 +589,7 @@ settrap(int sig, List l)
 {
     if (sig == -1)
         return 1;
-    if (jobbing && (sig == SIGTTOU || sig == SIGTSTP ||
-		    sig == SIGTTIN || sig == SIGPIPE)) {
+    if (jobbing && (sig == SIGTTOU || sig == SIGTSTP || sig == SIGTTIN)) {
         zerr("can't trap SIG%s in interactive shells", sigs[sig], 0);
         return 1;
     }
@@ -623,8 +623,7 @@ unsettrap(int sig)
     int trapped;
 
     if (sig == -1 || !(trapped = sigtrapped[sig]) ||
-	(jobbing && (sig == SIGTTOU || sig == SIGTSTP ||
-		     sig == SIGTTIN || sig == SIGPIPE))) {
+	(jobbing && (sig == SIGTTOU || sig == SIGTSTP || sig == SIGTTIN))) {
         return;
     }
     sigtrapped[sig] = 0;
@@ -654,15 +653,18 @@ unsettrap(int sig)
     }
 }
 
-/* Execute a trap function for a given signal */
+/* Execute a trap function for a given signal, possibly
+ * with non-standard sigtrapped & sigfuncs values
+ */
 
 /**/
 void
-dotrap(int sig)
+dotrapargs(int sig, int *sigtr, void *sigfn)
 {
     LinkList args;
     char *name, num[4];
     int trapret = 0;
+    int obreaks = breaks;
  
     /* if signal is being ignored or the trap function		      *
      * is NULL, then return					      *
@@ -672,14 +674,15 @@ dotrap(int sig)
      * intact without working too hard.  Special cases (e.g. calling  *
      * a trap for SIGINT after the error flag was set) are handled    *
      * by the calling code.  (PWS 1995/06/08).			      */
-    if ((sigtrapped[sig] & ZSIG_IGNORED) || !sigfuncs[sig] || errflag)
+    if ((*sigtr & ZSIG_IGNORED) || !sigfn || errflag)
         return;
 
-    sigtrapped[sig] |= ZSIG_IGNORED;
+    *sigtr |= ZSIG_IGNORED;
 
     lexsave();
     execsave();
-    if (sigtrapped[sig] & ZSIG_FUNC) {
+    breaks = 0;
+    if (*sigtr & ZSIG_FUNC) {
 	PERMALLOC {
 	    args = newlinklist();
 	    name = (char *) zalloc(5 + strlen(sigs[sig]));
@@ -689,11 +692,11 @@ dotrap(int sig)
 	    addlinknode(args, num);
 	} LASTALLOC;
 	trapreturn = -1;
-	doshfunc(sigfuncs[sig], args, 0, 1);
+	doshfunc(sigfn, args, 0, 1);
 	freelinklist(args, (FreeFunc) NULL);
 	zsfree(name);
     } else HEAPALLOC {
-	execlist(dupstruct(sigfuncs[sig]), 1, 0);
+	execlist(dupstruct(sigfn), 1, 0);
     } LASTALLOC;
     if (trapreturn > 0)
 	trapret = trapreturn;
@@ -705,9 +708,21 @@ dotrap(int sig)
     if (trapret > 0) {
 	breaks = loops;
 	errflag = 1;
+    } else {
+	breaks += obreaks;
+	if (breaks > loops)
+	    breaks = loops;
     }
 
-    if (sigtrapped[sig] != ZSIG_IGNORED)
-	sigtrapped[sig] &= ~ZSIG_IGNORED;
+    if (*sigtr != ZSIG_IGNORED)
+	*sigtr &= ~ZSIG_IGNORED;
+}
+
+/* Standard call to execute a trap for a given signal */
+
+/**/
+void
+dotrap(int sig)
+{
+    dotrapargs(sig, sigtrapped+sig, sigfuncs[sig]);
 }
- 
