@@ -1,5 +1,5 @@
 /*
- * $Id: subst.c,v 2.48 1996/07/28 22:29:59 hzoli Exp $
+ * $Id: subst.c,v 2.52 1996/10/15 20:16:35 hzoli Exp $
  *
  * subst.c - various substitutions
  *
@@ -10,9 +10,9 @@
  *
  * Permission is hereby granted, without written agreement and without
  * license or royalty fees, to use, copy, modify, and distribute this
- * software and its documentation for any purpose, provided that the
- * above copyright notice and the following two paragraphs appear in
- * all copies of this software.
+ * software and to distribute modified versions of this software for any
+ * purpose, provided that the above copyright notice and the following
+ * two paragraphs appear in all copies of this software.
  *
  * In no event shall Paul Falstad or the Zsh Development Group be liable
  * to any party for direct, indirect, special, incidental, or consequential
@@ -328,29 +328,28 @@ filesubstr(char **namptr, int assign)
     char *str = *namptr;
 
     if (*str == Tilde && str[1] != '=' && str[1] != Equals) {
+	char *ptr;
+	int val;
+
+	val = zstrtol(str + 1, &ptr, 10);
 	if (isend(str[1])) {   /* ~ */
 	    *namptr = dyncat(home, str + 1);
 	    return 1;
 	} else if (str[1] == '+' && isend(str[2])) {   /* ~+ */
 	    *namptr = dyncat(pwd, str + 2);
 	    return 1;
-	} else if (str[1] == '+' && idigit(str[2])) {   /* ~+42 */
-	    str++;
-	    goto dstack;
 	} else if (str[1] == '-' && isend(str[2])) {   /* ~- */
 	    char *tmp;
 	    *namptr = dyncat((tmp = oldpwd) ? tmp : pwd, str + 2);
 	    return 1;
-	} else if (str[1] == '-' && idigit(str[2])) {   /* ~-42 */
-	    str++;
-	    goto dstack;
-	} else if (idigit(str[1])) {   /* ~42 */
-	    char *ds, *ptr;
-	    int val;
-	dstack:
-	    val = zstrtol(str + 1, &ptr, 10);
-	    ds = dstackent(*str, val);
-	    if (!ds || !isend(*ptr))
+	} else if (!inblank(str[1]) && isend(*ptr) &&
+		   (!idigit(str[1]) || (ptr - str < 4))) {
+	    char *ds;
+
+	    if (val < 0)
+		val = -val;
+	    ds = dstackent(str[1], val);
+	    if (!ds)
 		return 0;
 	    *namptr = dyncat(ds, ptr);
 	    return 1;
@@ -655,6 +654,34 @@ get_strarg(char *s)
     return s;
 }
 
+/**/
+int
+get_intarg(char **s)
+{
+    char *t = get_strarg(*s + 1);
+    char *p, sav;
+    long ret;
+
+    if (!*t)
+	return -1;
+    sav = *t;
+    *t = '\0';
+    p = dupstring(*s + 2);
+    *s = t;
+    *t = sav;
+    if (parsestr(p))
+	return -1;
+    singsub(&p);
+    if (errflag)
+	return -1;
+    ret = matheval(p);
+    if (errflag)
+	return -1;
+    if (ret < 0)
+	ret = -ret;
+    return ret < 0 ? -ret : ret;
+}
+
 /* parameter substitution */
 
 #define	isstring(c) ((c) == '$' || (char)(c) == String || (char)(c) == Qstring)
@@ -709,7 +736,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 	inbrace = 1;
 	s++;
 	if (*s == '(' || *s == Inpar) {
-	    char *t, sav, *d;
+	    char *t, sav;
 	    int tt = 0;
 	    long num;
 	    int escapes = 0;
@@ -752,18 +779,8 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		    substr = 1;
 		    break;
 		case 'I':
-		    flnum = 0;
-		    t = get_strarg(++s);
-		    if (*t) {
-			sav = *t;
-			*t = '\0';
-			d = dupstring(s + 1);
-			untokenize(d);
-			if ((flnum = mathevalarg(s + 1, &d)) < 0)
-			    flnum = -flnum;
-			*t = sav;
-			s = t;
-		    } else
+		    flnum = get_intarg(&s);
+		    if (flnum < 0)
 			goto flagerr;
 		    break;
 
@@ -829,27 +846,17 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		    tt = 1;
 		/* fall through */
 		case 'r':
-		    t = get_strarg(++s);
-		    if (!*t)
+		    sav = s[1];
+		    num = get_intarg(&s);
+		    if (num < 0)
 			goto flagerr;
-		    sav = *t;
-		    *t = '\0';
-		    d = dupstring(s + 1);
-		    untokenize(d);
-		    if ((num = mathevalarg(d, &d)) < 0)
-			num = -num;
 		    if (tt)
 			prenum = num;
 		    else
 			postnum = num;
-		    *t = sav;
-		    sav = *s;
-		    s = t + 1;
-		    if (*s != sav) {
-			s--;
+		    if (s[1] != sav)
 			break;
-		    }
-		    t = get_strarg(s);
+		    t = get_strarg(++s);
 		    if (!*t)
 			goto flagerr;
 		    sav = *t;
@@ -1137,6 +1144,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 	case Equals:
 	    if (vunset) {
 		char sav = *idend;
+		int l;
 
 		*idend = '\0';
 		val = dupstring(u);
@@ -1149,16 +1157,29 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		    char *arr[2], **t, **a, **p;
 		    if (spsep || spbreak) {
 			aval = sepsplit(val, spsep, 0);
-			isarr = 1;
+			isarr = 2;
 			sep = spsep = NULL;
 			spbreak = 0;
+			l = arrlen(aval);
+			if (l && !*(aval[l-1]))
+			    l--;
+			if (l && !**aval)
+			    l--, t = aval + 1;
+			else
+			    t = aval;
 		    } else if (!isarr) {
 			arr[0] = val;
 			arr[1] = NULL;
-			aval = arr;
+			t = aval = arr;
+			l = 1;
+		    } else
+			l = arrlen(aval), t = aval;
+		    p = a = zalloc(sizeof(char *) * (l + 1));
+		    while (l--) {
+			untokenize(*t);
+			*p++ = ztrdup(*t++);
 		    }
-		    p = a = zcalloc(sizeof(char *) * (arrlen(aval) + 1));
-		    for (t = aval; *t; untokenize(*t), *p++ = ztrdup(*t++));
+		    *p++ = NULL;
 		    setaparam(idbeg, a);
 		} else {
 		    untokenize(val);
@@ -1312,7 +1333,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 	    else if (!aval[1])
 		val = aval[0];
 	    else
-		isarr = 1;
+		isarr = 2;
 	}
     }
     if (casmod) {
@@ -1406,7 +1427,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		for (tn = firstnode(tl); tn; incnode(tn)) {
 		    strcatsub(&y, ostr, aptr, x, xlen,
 			      (char *) getdata(tn), globsubst);
-		    if (qt && !*y)
+		    if (qt && !*y && isarr != 2)
 			y = dupstring(nulstring);
 		    if (i == 1)
 			setdata(n, (void *) y);
@@ -1423,7 +1444,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		return NULL;
 	    xlen = strlen(x);
 	    strcatsub(&y, ostr, aptr, x, xlen, NULL, globsubst);
-	    if (qt && !*y)
+	    if (qt && !*y && isarr != 2)
 		y = dupstring(nulstring);
 	    setdata(n, (void *) y);
 
@@ -1436,7 +1457,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 				  premul, postmul);
 		if (eval && parsestr(x))
 		    return NULL;
-		if (qt && !*x)
+		if (qt && !*x && isarr != 2)
 		    y = dupstring(nulstring);
 		else if (globsubst)
 		    tokenize(y = dupstring(x));
@@ -1453,7 +1474,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		return NULL;
 	    xlen = strlen(x);
 	    *str = strcatsub(&y, aptr, aptr, x, xlen, s, globsubst);
-	    if (qt && !*y)
+	    if (qt && !*y && isarr != 2)
 		y = dupstring(nulstring);
 	    insertlinknode(l, n, (void *) y), incnode(n);
 	}
@@ -1472,7 +1493,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 	    return NULL;
 	xlen = strlen(x);
 	*str = strcatsub(&y, ostr, aptr, x, xlen, s, globsubst);
-	if (qt && !*y)
+	if (qt && !*y && isarr != 2)
 	    y = dupstring(nulstring);
 	setdata(n, (void *) y);
     }
@@ -1568,6 +1589,7 @@ modify(char **str, char **ptr)
 		for (tt = hsubl; *tt; tt++)
 		    if (INULL(*tt))
 			chuck(tt--);
+		untokenize(hsubl);
 		for (tt = hsubr = ztrdup(ptr2); *tt; tt++)
 		    if (INULL(*tt))
 			chuck(tt--);
@@ -1608,18 +1630,8 @@ modify(char **str, char **ptr)
 		(*ptr)++;
 		break;
 	    case 'F':
-		rec = -1;
+		rec = get_intarg(ptr);
 		(*ptr)++;
-		ptr1 = get_strarg(ptr2 = *ptr);
-		if ((sav = *ptr1))
-		    *ptr1 = '\0';
-		ptr2 = dupstring(ptr2 + 1);
-		if (sav)
-		    *ptr1 = sav;
-		untokenize(ptr2);
-		rec = mathevalarg(ptr2, &ptr2);
-		*ptr = ptr1 + 1;
-		c = '\0';
 		break;
 	    default:
 		*ptr = lptr;
