@@ -231,7 +231,11 @@ struct timezone {
 # else
 #  include <termios.h>
 # endif
-# define VDISABLEVAL 0
+# ifdef _POSIX_VDISABLE
+#  define VDISABLEVAL _POSIX_VDISABLE
+# else
+#  define VDISABLEVAL 0
+# endif
 # define HAS_TIO 1
 #else    /* not TERMIOS */
 # ifdef HAVE_TERMIO_H
@@ -247,7 +251,7 @@ struct timezone {
 #include <termcap.h>
 #endif
 
-#ifdef GWINSZ_IN_SYS_IOCTL
+#if defined(GWINSZ_IN_SYS_IOCTL) || defined(CLOBBERS_TYPEAHEAD)
 # include <sys/ioctl.h>
 #endif
 #ifdef WINSIZE_IN_PTEM
@@ -377,19 +381,6 @@ struct timezone {
 # define RLIMIT_VMEM RLIMIT_AS
 #endif
 
-/* RLIM_T is the type for system calls involving resource limits     */
-/* Eventually the following will be replaced by a check in configure */
-
-#if defined(BSD4_4) && (BSD > 199300)
-# define RLIM_T_IS_QUAD_T
-#endif
-
-#ifdef RLIM_T_IS_QUAD_T
-# define RLIM_T quad_t
-#else
-# define RLIM_T long
-#endif
-
 /* DIGBUFSIZ is the length of a buffer which can hold the -LONG_MAX-1 *
  * converted to printable decimal form including the sign and the     *
  * terminating null character. Below 0.30103 > lg 2.                  */
@@ -465,7 +456,7 @@ struct timezone {
 #endif
 
 #ifndef HAVE_LSTAT
-# define lstat stat
+# define lstat(X,Y) stat(X,Y)
 #endif
 
 #ifndef F_OK          /* missing macros for access() */
