@@ -274,9 +274,7 @@ slashsplit(char *s)
     q = r = (char **) zalloc(sizeof(char **) * (t0 + 2));
 
     while ((t = strchr(s, '/'))) {
-	*t = '\0';
-	*q++ = ztrdup(s);
-	*t = '/';
+	*q++ = ztrduppfx(s, t - s);
 	while (*t == '/')
 	    t++;
 	if (!*t) {
@@ -792,11 +790,6 @@ gettyinfo(struct ttyinfo *ti)
 	ioctl(SHTTY, TIOCGETC, &ti->tchars);
 	ioctl(SHTTY, TIOCGLTC, &ti->ltchars);
 # endif
-#endif
-#ifdef TIOCGWINSZ
-/*	if (ioctl(SHTTY, TIOCGWINSZ, &ti->winsize) == -1)
-	    	zerr("bad tiocgwinsz: %e",NULL,errno);*/
-	ioctl(SHTTY, TIOCGWINSZ, (char *)&ti->winsize);
 #endif
     }
 }
@@ -848,24 +841,13 @@ adjustwinsize(void)
 	return;
 
     ioctl(SHTTY, TIOCGWINSZ, (char *)&shttyinfo.winsize);
-    if (shttyinfo.winsize.ws_col)
-	columns = shttyinfo.winsize.ws_col;
-    if (shttyinfo.winsize.ws_row)
-	lines = shttyinfo.winsize.ws_row;
-    if (oldcols != columns) {
-	if (columns < 2)
-	    opts[USEZLE] = 0;
-	if (lines < 2)
-	    opts[SINGLELINEZLE] = 1;
-	if (zleactive) {
-	    resetneeded = winchanged = 1;
-	    refresh();
-	}
-	setintenv("COLUMNS", columns);
+    setiparam("COLUMNS", shttyinfo.winsize.ws_col);
+    setiparam("LINES", shttyinfo.winsize.ws_row);
+    if (zleactive && (oldcols != columns || oldrows != lines)) {
+	resetneeded = winchanged = 1;
+	refresh();
     }
-    if (oldrows != lines)
-	setintenv("LINES", lines);
-#endif   /*  TIOCGWINSZ */
+#endif   /* TIOCGWINSZ */
 }
 
 /* Move a fd to a place >= 10 and mark the new fd in fdtable.  If the fd *
@@ -1093,16 +1075,13 @@ zstrtol(const char *s, char **t, int base)
 
 /* Convert string to quad_t. */
 
-#ifdef RLIM_T_IS_QUAD_T
-/* The prototype of this function will be included in utils.pro even if *
- * RLIM_T_IS_QUAD_T is undefined so we cannot use quad_t prototype here *
- * since not all system have it.                                        */
+#if defined(RLIM_T_IS_QUAD_T) || defined(RLIM_T_IS_UNSIGNED)
 
 /**/
-RLIM_T
-zstrtoq(const char *s, char **t, int base)
+rlim_t
+zstrtorlimit(const char *s, char **t, int base)
 {
-    quad_t ret = 0;
+    rlim_t ret = 0;
  
     if (!base)
 	if (*s != '0')
@@ -1226,7 +1205,7 @@ spckword(char **s, int hist, int cmd, int ask)
     int ne;
     int preflen = 0;
 
-    if (**s == '-' || **s == '%')
+    if ((histdone & HISTFLAG_NOEXEC) || **s == '-' || **s == '%')
 	return;
     if (!strcmp(*s, "in"))
 	return;
@@ -1513,11 +1492,14 @@ spacesplit(char *s, int allownull)
 {
     char *t, **ret, **ptr;
 
-    skipwsep(&s);
-    ptr = ret = (char **) ncalloc(sizeof(*ret) * (wordcount(s, NULL, 0) + 1));
+    ptr = ret = (char **) ncalloc(sizeof(*ret) * (wordcount(s, NULL, -!allownull) + 1));
 
+    t = s;
+    skipwsep(&s);
     if (*s && isep(*s == Meta ? s[1] ^ 32 : *s))
 	*ptr++ = dupstring(allownull ? "" : nulstring);
+    else if (!allownull && t != s)
+	*ptr++ = dupstring("");
     while (*s) {
 	if (isep(*s == Meta ? s[1] ^ 32 : *s)) {
 	    if (*s == Meta)
@@ -1532,8 +1514,11 @@ spacesplit(char *s, int allownull)
 	    ztrncpy(*ptr++, t, s - t);
 	} else
 	    *ptr++ = dupstring(nulstring);
+	t = s;
 	skipwsep(&s);
     }
+    if (!allownull && t != s)
+	*ptr++ = dupstring("");
     *ptr = NULL;
     return ret;
 }
@@ -1620,25 +1605,29 @@ wordcount(char *s, char *sep, int mul)
 	    if ((c && *(s + sl)) || mul)
 		r++;
     } else {
-	char *t;
+	char *t = s;
 
 	r = 0;
-	if (!mul)
+	if (mul <= 0)
 	    skipwsep(&s);
-	if (*s && isep(*s == Meta ? s[1] ^ 32 : *s))
+	if ((*s && isep(*s == Meta ? s[1] ^ 32 : *s)) ||
+	    (mul < 0 && t != s))
 	    r++;
-	for (t = s; *t; r++) {
-	    if (isep(*t == Meta ? t[1] ^ 32 : *t)) {
-		if (*t == Meta)
-		    t++;
-		t++;
-		if (!mul)
-		    skipwsep(&t);
+	for (; *s; r++) {
+	    if (isep(*s == Meta ? s[1] ^ 32 : *s)) {
+		if (*s == Meta)
+		    s++;
+		s++;
+		if (mul <= 0)
+		    skipwsep(&s);
 	    }
-	    findsep(&t, NULL);
-	    if (!mul)
-		skipwsep(&t);
+	    findsep(&s, NULL);
+	    t = s;
+	    if (mul <= 0)
+		skipwsep(&s);
 	}
+	if (mul < 0 && t != s)
+	    r++;
     }
     return r;
 }
@@ -3375,6 +3364,28 @@ dosetopt(int optno, int value, int force)
     return 0;
 }
 
+/**/
+char *
+dupstrpfx(const char *s, int len)
+{
+    char *r = ncalloc(len + 1);
+
+    memcpy(r, s, len);
+    r[len] = '\0';
+    return r;
+}
+
+/**/
+char *
+ztrduppfx(const char *s, int len)
+{
+    char *r = zalloc(len + 1);
+
+    memcpy(r, s, len);
+    r[len] = '\0';
+    return r;
+}
+
 #ifdef DEBUG
 
 /**/
