@@ -1,5 +1,5 @@
 /*
- * $Id: zle_hist.c,v 2.23 1996/07/28 20:48:29 hzoli Exp $
+ * $Id: zle_hist.c,v 2.24 1996/10/15 20:16:35 hzoli Exp $
  *
  * zle_hist.c - history editing
  *
@@ -10,9 +10,9 @@
  *
  * Permission is hereby granted, without written agreement and without
  * license or royalty fees, to use, copy, modify, and distribute this
- * software and its documentation for any purpose, provided that the
- * above copyright notice and the following two paragraphs appear in
- * all copies of this software.
+ * software and to distribute modified versions of this software for any
+ * purpose, provided that the above copyright notice and the following
+ * two paragraphs appear in all copies of this software.
  *
  * In no event shall Paul Falstad or the Zsh Development Group be liable
  * to any party for direct, indirect, special, incidental, or consequential
@@ -483,8 +483,17 @@ insertlastword(void)
 	feep();
 	return;
     }
-    s = he->text + he->words[2*he->nwords-2];
-    t = he->text + he->words[2*he->nwords-1];
+    if (zmult > 0) {
+	zmult = he->nwords - (zmult - 1);
+    } else {
+	zmult = 1 - zmult;
+    }
+    if (zmult < 1 || zmult > he->nwords) {
+	feep();
+	return;
+    }
+    s = he->text + he->words[2*zmult-2];
+    t = he->text + he->words[2*zmult-1];
     save = *t;
     *t = '\0';			/* ignore trailing whitespace */
 
