@@ -370,7 +370,7 @@ zleread(char *lp, char *rp)
 	/* We could be smarter and default to a system read. */
 
 	/* If we just got a new shout, make sure the terminal is set up. */
-	if (!termok)
+	if (termflags & TERM_UNKNOWN)
 	    init_term();
     }
 
@@ -428,7 +428,7 @@ zleread(char *lp, char *rp)
 
 	    statusline = NULL;
 	    bindk = getkeycmd();
-	    if (!ll && c == eofchar) {
+	    if (!ll && isfirstln && c == eofchar) {
 		eofsent = 1;
 		break;
 	    }
@@ -1109,7 +1109,7 @@ describekeybriefly(void)
     if (cmd < 0)
 	return;
     trashzle();
-    clearflag = (isset(USEZLE) && termok &&
+    clearflag = (isset(USEZLE) && !termflags &&
 		 (isset(ALWAYSLASTPROMPT) && zmult == 1)) ||
 	(unset(ALWAYSLASTPROMPT) && zmult != 1);
     printbind(keybuf, (len = strlen(keybuf)) ? len : 1, 0);
@@ -1156,7 +1156,7 @@ whereis(void)
 	return;
     funcfound = 0;
     trashzle();
-    clearflag = (isset(USEZLE) && termok &&
+    clearflag = (isset(USEZLE) && !termflags &&
 		 (isset(ALWAYSLASTPROMPT) && zmult == 1)) ||
 	(unset(ALWAYSLASTPROMPT) && zmult != 1);
     if (func == z_selfinsert || func == z_undefinedkey)
