@@ -1,5 +1,5 @@
 /*
- * $Id: zle_refresh.c,v 2.13 1996/08/12 01:39:05 hzoli Exp $
+ * $Id: zle_refresh.c,v 2.19 1996/10/15 20:16:35 hzoli Exp $
  *
  * zle_refresh.c - screen update
  *
@@ -10,9 +10,9 @@
  *
  * Permission is hereby granted, without written agreement and without
  * license or royalty fees, to use, copy, modify, and distribute this
- * software and its documentation for any purpose, provided that the
- * above copyright notice and the following two paragraphs appear in
- * all copies of this software.
+ * software and to distribute modified versions of this software for any
+ * purpose, provided that the above copyright notice and the following
+ * two paragraphs appear in all copies of this software.
  *
  * In no event shall Paul Falstad or the Zsh Development Group be liable
  * to any party for direct, indirect, special, incidental, or consequential
@@ -34,8 +34,12 @@
 
 #ifdef HAVE_SELECT
 #define SELECT_ADD_COST(X)	cost += X
+#define zputc(a, b)		putc(a, b), cost++
+#define zwrite(a, b, c, d)	fwrite(a, b, c, d), cost += (b * c)
 #else
 #define SELECT_ADD_COST(X)
+#define zputc(a, b)		putc(a, b)
+#define zwrite(a, b, c, d)	fwrite(a, b, c, d)
 #endif
 
 /* Oct/Nov 94: <mason> some code savagely redesigned to fix several bugs -
@@ -54,7 +58,7 @@ static int more_start,		/* more text before start of screen?	    */
     pptw, rpw,                  /* prompt widths on screen                  */
     vcs, vln,			/* video cursor position column & line	    */
     vmaxln,			/* video maximum number of lines	    */
-    winw, winh,			/* window width & height		    */
+    winw, winh, rwinh,		/* window width & height		    */
     winpos;			/* singlelinezle: line's position in window */
 static unsigned pmpt_attr = 0,	/* text attributes after displaying prompt  */
     rpmpt_attr = 0;		/* text attributes after displaying rprompt */
@@ -66,25 +70,27 @@ resetvideo(void)
     int ln;
     static int lwinw = -1, lwinh = -1;	/* last window width & height */
  
-    winw = COLUMNS;  /* terminal width */
-    if (isset(SINGLELINEZLE) || termok != TERM_OK)
+    genprompts();
+    winw = columns;  /* terminal width */
+    if (termflags & TERM_SHORT)
 	winh = 1;
     else
 	winh = (lines < 2) ? 24 : lines;
+    rwinh = lines;		/* keep the real number of lines */
     winpos = vln = vmaxln = 0;
     if (lwinw != winw || lwinh != winh) {
 	if (nbuf) {
 	    for (ln = 0; ln != lwinh; ln++) {
-		zfree(nbuf[ln], lwinw + 1);
-		zfree(obuf[ln], lwinw + 1);
+		zfree(nbuf[ln], lwinw + 2);
+		zfree(obuf[ln], lwinw + 2);
 	    }
 	    free(nbuf);
 	    free(obuf);
 	}
 	nbuf = (char **)zcalloc((winh + 1) * sizeof(char *));
 	obuf = (char **)zcalloc((winh + 1) * sizeof(char *));
-	nbuf[0] = (char *)zalloc(winw + 1);
-	obuf[0] = (char *)zalloc(winw + 1);
+	nbuf[0] = (char *)zalloc(winw + 2);
+	obuf[0] = (char *)zalloc(winw + 2);
 
 	lwinw = winw;
 	lwinh = winh;
@@ -109,73 +115,74 @@ resetvideo(void)
 }
 
 /*
- * Jul 96: <mason> changed to single line scroll for higher speed terminals
- *  I've seperated the loops for readability (and it's slightly faster)
- *  Returns line number to be used next
+ * Nov 96: <mason> changed to single line scroll
  */
 
 /**/
-int
+void
 scrollwindow(int tline)
 {
-    int t0, hwinh;
+    int t0;
     char *s;
 
-    if (tline || baud >= 19200) {	/* single line scroll */
-	hwinh = 1;
-	s = nbuf[tline];
-	for (t0 = tline; t0 < winh - 1; t0++)
-	    nbuf[t0] = nbuf[t0 + 1];
-	nbuf[winh - 1] = s;
-    } else { 				/* half screen scroll */
-	hwinh = winh / 2;
-	for (t0 = 0; t0 != winh - hwinh; t0++) {
-	    s = nbuf[t0];
-	    nbuf[t0] = nbuf[t0 + hwinh];
-	    nbuf[t0 + hwinh] = s;
-	}
-    }
+    s = nbuf[tline];
+    for (t0 = tline; t0 < winh - 1; t0++)
+	nbuf[t0] = nbuf[t0 + 1];
+    nbuf[winh - 1] = s;
     if (!tline)
 	more_start = 1;
-    return winh - hwinh;
+    return;
 }
 
 /* this is the messy part. */
 /* this define belongs where it's used!!! */
 
-#define nextline				\
-{						\
-    *s = '\0';					\
-    if (++ln == winh)				\
-	if (nvln != --ln && nvln != -1)		\
-	    break;				\
-	else {					\
-	    ln = scrollwindow(0);		\
-	    if(nvln != -1)			\
-		nvln -= winh - ln;		\
-	}					\
-    if (!nbuf[ln])				\
-	nbuf[ln] = (char *)zalloc(winw + 1);	\
-    s = (unsigned char *)nbuf[ln];		\
-    sen = s + winw;				\
+#define nextline					\
+{							\
+    *s = '\0';						\
+    if (ln != winh - 1)					\
+	ln++;						\
+    else {						\
+	if (!canscroll)	{				\
+	    if (nvln != -1 && nvln != winh - 1		\
+		&& (numscrolls != onumscrolls - 1	\
+		    || nvln <= winh / 2))		\
+	        break;					\
+	    numscrolls++;				\
+	    canscroll = winh / 2;			\
+	}						\
+	canscroll--;					\
+	scrollwindow(0);				\
+	if (nvln != -1)					\
+	    nvln--;					\
+    }							\
+    if (!nbuf[ln])					\
+	nbuf[ln] = (char *)zalloc(winw + 2);		\
+    s = (unsigned char *)nbuf[ln];			\
+    sen = s + winw;					\
 }
 
-#define snextline				\
-{						\
-    *s = '\0';					\
-    if (++ln == winh)				\
-	if (tosln <= nvln + 1) {		\
-	    ln = scrollwindow(0);		\
-	    if (nvln)				\
-		nvln--, tosln--;		\
-	} else {				\
-	    tosln--;				\
-	    ln = scrollwindow(tosln);		\
-	}					\
-    if (!nbuf[ln])				\
-	nbuf[ln] = (char *)zalloc(winw + 1);	\
-    s = (unsigned char *)nbuf[ln];		\
-    sen = s + winw;				\
+#define snextline					\
+{							\
+    *s = '\0';						\
+    if (ln != winh - 1)					\
+	ln++;						\
+    else						\
+	if (tosln < 3) {				\
+	    more_status = 1;				\
+	    scrollwindow(tosln + 1);			\
+	} else if (tosln - 1 <= nvln) {			\
+	    scrollwindow(0);				\
+	    if (nvln)					\
+		nvln--, tosln--;			\
+	} else {					\
+	    tosln--;					\
+	    scrollwindow(tosln);			\
+	}						\
+    if (!nbuf[ln])					\
+	nbuf[ln] = (char *)zalloc(winw + 2);		\
+    s = (unsigned char *)nbuf[ln];			\
+    sen = s + winw;					\
 }
 
 #ifdef TIOCGWINSZ
@@ -186,7 +193,9 @@ int cleareol,			/* clear to end-of-line (if can't cleareod) */
     clearf = 0,			/* alwayslastprompt used immediately before */
     hasam;			/* terminal should have auto-margin	    */
 static int put_rpmpt,		/* whether we should display right-prompt   */
-    oput_rpmpt;			/* whether displayed right-prompt last time */
+    oput_rpmpt,			/* whether displayed right-prompt last time */
+    oxtabs,			/* oxtabs - tabs expand to spaces if set    */
+    numscrolls, onumscrolls;
 extern int clearflag;		/* set to non-zero if alwayslastprompt used */
 
 /**/
@@ -194,7 +203,9 @@ void
 refresh(void)
 {
     static int inlist;		/* avoiding recursion                        */
-    int ln = 0,			/* current line we're working on	     */
+    int canscroll = 0,		/* number of lines we are allowed to scroll  */
+	ln = 0,			/* current line we're working on	     */
+	more_status = 0,	/* more stuff in status line		     */
 	nvcs = 0, nvln = -1,	/* video cursor column and line		     */
 	t0 = -1,		/* tmp					     */
 	tosln = 0;		/* tmp in statusline stuff		     */
@@ -214,9 +225,20 @@ refresh(void)
 #ifdef HAVE_SELECT
     cost = 0;			/* reset */
 #endif
+
+/* Nov 96: <mason>  I haven't checked how complete this is.  sgtty stuff may
+   or may not work */
+    oxtabs = ((SGTTYFLAG & SGTABTYPE) == SGTABTYPE);
+
     cleareol = 0;		/* unset */
     more_start = more_end = 0;	/* unset */
+    if (isset(SINGLELINEZLE) || lines < 3
+	|| (termflags & (TERM_NOUP | TERM_BAD | TERM_UNKNOWN)))
+	termflags |= TERM_SHORT;
+    else
+	termflags &= ~TERM_SHORT;
     if (resetneeded) {
+	onumscrolls = 0;
 	setterm();
 #ifdef TIOCGWINSZ
 	if (winchanged) {
@@ -228,6 +250,12 @@ refresh(void)
 	resetvideo();
 	resetneeded = 0;	/* unset */
 	oput_rpmpt = 0;		/* no right-prompt currently on screen */
+
+	/* we probably should only have explicitly set attributes */
+	tsetcap(TCALLATTRSOFF, 0);
+	tsetcap(TCSTANDOUTEND, 0);
+	tsetcap(TCUNDERLINEEND, 0);
+
         if (!clearflag)
             if (tccan(TCCLEAREOD))
                 tcout(TCCLEAREOD);
@@ -235,24 +263,24 @@ refresh(void)
                 cleareol = 1;   /* request: clear to end of line */
         if (t0 > -1)
             olnct = t0;
-        if (isset(SINGLELINEZLE) || termok != TERM_OK)
+        if (termflags & TERM_SHORT)
             vcs = 0;
-        else if (!clearflag && lpptlen) {
-            fwrite(lpptbuf, lpptlen, 1, shout);
-	    fflush(shout);
-	}
+        else if (!clearflag && lpptlen)
+            zwrite(lpptbuf, lpptlen, 1, shout);
 	if (clearflag) {
-	    putc('\r', shout);
+	    zputc('\r', shout);
 	    vcs = 0;
 	    moveto(0, pptw);
 	}
+	fflush(shout);
 	clearf = clearflag;
-    } else if (winw != COLUMNS)
+    } else if (winw != columns || rwinh != lines)
 	resetvideo();
 
-/* now winw equals columns; now all width comparisons can be made to winw */
+/* now winw equals columns and winh equals lines 
+   width comparisons can be made with winw, height comparisons with winh */
 
-    if (isset(SINGLELINEZLE) || termok != TERM_OK) {
+    if (termflags & TERM_SHORT) {
 	singlerefresh();
 	return;
     }
@@ -265,13 +293,14 @@ refresh(void)
 	cs = 0;
     }
     scs = line + cs;
+    numscrolls = 0;
 
 /* first, we generate the video line buffers so we know what to put on
    the screen - also determine final cursor position (nvln, nvcs) */
 
     /* Deemed necessary by PWS 1995/05/15 due to kill-line problems */
     if (!*nbuf)
-	*nbuf = (char *)zalloc(winw + 1);
+	*nbuf = (char *)zalloc(winw + 2);
 
     s = (unsigned char *)(nbuf[ln = 0] + pptw);
     t = line;
@@ -280,33 +309,41 @@ refresh(void)
 	if (t == scs)			/* if cursor is here, remember it */
 	    nvcs = s - (unsigned char *)(nbuf[nvln = ln]);
 
-	if (*t == '\n')			/* newline */
+	if (*t == '\n')	{		/* newline */
+	    nbuf[ln][winw + 1] = '\0';	/* text not wrapped */
 	    nextline
-	else if (*t == '\t') {		/* tab */
+	} else if (*t == '\t') {		/* tab */
 	    t0 = (char *)s - nbuf[ln];
-	    if ((t0 | 7) + 1 >= winw)
+	    if ((t0 | 7) + 1 >= winw) {
+		nbuf[ln][winw + 1] = '\n';	/* text wrapped */
 		nextline
-	    else
+	    } else
 		do
 		    *s++ = ' ';
 		while ((++t0) & 7);
 	} else if (icntrl(*t)) {	/* other control character */
 	    *s++ = '^';
-	    if (s == sen)
+	    if (s == sen) {
+		nbuf[ln][winw + 1] = '\n';	/* text wrapped */
 		nextline
+	    }
 	    *s++ = (*t == 127) ? '?' : (*t | '@');
 	} else				/* normal character */
 	    *s++ = *t;
-	if (s == sen)
+	if (s == sen) {
+	    nbuf[ln][winw + 1] = '\n';	/* text wrapped */
 	    nextline
+	}
     }
 
 /* if we're really on the next line, don't fake it; do everything properly */
     if (t == scs && (nvcs = s - (unsigned char *)(nbuf[nvln = ln])) == winw) {
+	nbuf[ln][winw + 1] = '\n';	/* text wrapped */
+	switch ('\0') { 	/* a sad hack to make the break */
+	case '\0':		/* in nextline work */
+	    nextline
+	}
 	*s = '\0';
-	if (!nbuf[++ln])
-	    nbuf[ln] = (char *)zalloc(winw + 1);
-	s = (unsigned char *)nbuf[ln];
 	nvcs = 0;
 	nvln++;
     }
@@ -316,18 +353,30 @@ refresh(void)
 
     if (statusline) {
 	tosln = ln + 1;
+        if (ln == winh - 1) {
+	    if (nvln > 0) {
+		scrollwindow(0);
+		nvln--;
+	    }
+	    tosln--;
+	}
+	nbuf[ln][winw + 1] = '\0';	/* text not wrapped */
 	snextline
 	t = (unsigned char *)statusline;
-	for (; t < (unsigned char *)statusline+statusll; t++) {
+	for (; t < (unsigned char *)statusline + statusll; t++) {
 	    if (icntrl(*t)) {	/* simplified processing in the status line */
 		*s++ = '^';
-		if (s == sen)
+		if (s == sen) {
+		    nbuf[ln][winw + 1] = '\n';	/* text wrapped */
 		    snextline
+		}
 		*s++ = (*t == 127) ? '?' : (*t | '@');
 	    } else
 		*s++ = *t;
-	    if (s == sen)
+	    if (s == sen) {
+		nbuf[ln][winw + 1] = '\n';	/* text wrapped */
 		snextline
+	    }
 	}
     }
 
@@ -336,13 +385,19 @@ refresh(void)
 	if (!statusline)
 	    tosln = winh;
 	strncpy(nbuf[tosln - 1] + winw - 7, " <.... ", 7);
-	nbuf[tosln - 1][winw] = '\0';
+	nbuf[tosln - 1][winw] = nbuf[tosln - 1][winw + 1] = '\0';
+    }
+
+/* insert <....> at end of first status line if status is too big */
+    if (more_status) {
+	strncpy(nbuf[tosln] + winw - 8, " <....> ", 8);
+	nbuf[tosln][winw] = nbuf[tosln][winw + 1] = '\0';
     }
 
     *s = '\0';
     nlnct = ln + 1;
     for (ln = nlnct; ln < winh; ln++)
-	zfree(nbuf[ln], winw + 1), nbuf[ln] = NULL;
+	zfree(nbuf[ln], winw + 2), nbuf[ln] = NULL;
 
 /* determine whether the right-prompt exists and can fit on the screen */
     if (!more_start)
@@ -354,7 +409,7 @@ refresh(void)
 	t0 = t0 > 5 ? 5 : t0;
 	strncpy(nbuf[0] + pptw, ">....", t0);
 	memset(nbuf[0] + pptw + t0, ' ', winw - t0 - pptw);
-	nbuf[0][winw] = '\0';
+	nbuf[0][winw] = nbuf[0][winw + 1] = '\0';
     }
 
     for (ln = 0; !clearf && (ln < nlnct); ln++) {
@@ -372,7 +427,7 @@ refresh(void)
 		nbuf[ln] && !strncmp(nbuf[ln], obuf[ln + 1], 16)) {
 		moveto(ln, 0);
 		tcout(TCDELLINE);
-		zfree(obuf[ln], winw + 1);
+		zfree(obuf[ln], winw + 2);
 		for (t0 = ln; t0 != olnct; t0++)
 		    obuf[t0] = obuf[t0 + 1];
 		obuf[--olnct] = NULL;
@@ -398,7 +453,7 @@ refresh(void)
     /* output the right-prompt if appropriate */
 	if (put_rpmpt && !ln && !oput_rpmpt) {
 	    moveto(0, winw - 1 - rpw);
-	    fwrite(rpptbuf, rpptlen, 1, shout);
+	    zwrite(rpptbuf, rpptlen, 1, shout);
 	    vcs = winw - 1;
 	/* reset character attributes to that set by the main prompt */
 	    txtchange = pmpt_attr;
@@ -456,6 +511,7 @@ individually */
     ovln = nvln;
     olnct = nlnct;
     oput_rpmpt = put_rpmpt;
+    onumscrolls = numscrolls;
     if (nlnct > vmaxln)
 	vmaxln = nlnct;
     fflush(shout);		/* make sure everything is written out */
@@ -490,29 +546,40 @@ refreshline(int ln)
 	char_ins = 0,		/* number of characters inserted/deleted */
 	col_cleareol,		/* clear to end-of-line from this column */
 	i, j,			/* tmp					 */
-	nllen, ollen;		/* new and old line buffer lengths	 */
+	ins_last,		/* insert pushed last character off line */
+	nllen, ollen,		/* new and old line buffer lengths	 */
+	rnllen;			/* real new line buffer length		 */
 
 /* 0: setup */
     nl = nbuf[ln];
-    nllen = nl ? strlen(nl) : 0;
+    rnllen = nllen = nl ? strlen(nl) : 0;
     ol = obuf[ln] ? obuf[ln] : "";
     ollen = strlen(ol);
 
+/* optimisation: can easily happen for clearing old lines.  If the terminal has
+   the capability, then this is the easiest way to skip unnecessary stuff */
+    if (cleareol && !nllen && !(hasam && ln < nlnct - 1)
+	&& tccan(TCCLEAREOL)) {
+	moveto(ln, 0);
+	tcout(TCCLEAREOL);
+	return;	
+    }
+
 /* 1: pad out the new buffer with spaces to contain _all_ of the characters
       which need to be written. do this now to allow some pre-processing */
 
     if (cleareol 		/* request to clear to end of line */
 	|| !nllen 		/* no line buffer given */
 	|| (ln == 0 && (put_rpmpt != oput_rpmpt))) {	/* prompt changed */
-	p1 = halloc(winw + 1);
+	p1 = halloc(winw + 2);
 	if (nllen)
 	    strncpy(p1, nl, nllen);
 	memset(p1 + nllen, ' ', winw - nllen);
-	p1[winw] = '\0';
+	p1[winw] = p1[winw + 1] = '\0';
 	if (ln && nbuf[ln])
-	    strncpy(nl, p1, winw + 1);	/* next time obuf will be up-to-date */
+	    strncpy(nl, p1, winw + 2);	/* next time obuf will be up-to-date */
 	else
-	    nl = p1;		/* shouldn't happen */
+	    nl = p1;		/* don't keep the padding for prompt line */
 	nllen = winw;
     } else if (ollen > nllen) { /* make new line at least as long as old */
 	p1 = halloc(ollen + 1);
@@ -528,7 +595,7 @@ refreshline(int ln)
    With automatic margins, we shouldn't do it if there is another line, in
    case it messes up cut and paste. */
 
-    if (hasam && ln < nlnct - 1)
+    if (hasam && ln < nlnct - 1 && rnllen == winw)
 	col_cleareol = -2;	/* clearing eol would be evil so don't */
     else {
 	col_cleareol = -1;
@@ -544,28 +611,58 @@ refreshline(int ln)
 /* 2b: first a new trick for automargin niceness - good for cut and paste */
 
     if (hasam && vcs == winw) {
-	vln++, vcs = 1;
-	putc(*nbuf[vln], shout);
-	nl++;
-	if (*ol)
-	    ol++;
-	ccs = 1;
+	if (nbuf[vln] && nbuf[vln][vcs + 1] == '\n') {
+	    vln++, vcs = 1;
+            if (nbuf[vln]  && *nbuf[vln])
+		zputc(*nbuf[vln], shout);
+	    else
+		zputc(' ', shout);  /* I don't think this should happen */
+	    if (ln == vln) {	/* better safe than sorry */
+		nl++;
+		if (*ol)
+		    ol++;
+		ccs = 1;
+	    }			/* else  hmmm... I wonder what happened */
+	} else {
+	    vln++, vcs = 0;
+	    zputc('\n', shout);
+	}
+    }
+    ins_last = 0;
+
+/* 2c: if we're on the first line, start checking at the end of the prompt;
+   we shouldn't be doing anything within the prompt */
+
+    if (ln == 0 && pptw) {
+	i = pptw - ccs;
+	j = strlen(ol);
+	nl += i;
+	ol += (i > j ? j : i);	/* if ol is too short, point it to '\0' */
+	ccs = pptw;
     }
 
 /* 3: main display loop - write out the buffer using whatever tricks we can */
 
     for (;;) {
-	if (*nl && nl[1] == ol[1])	/* skip only if second chars match */
+	if (*nl && *ol && nl[1] == ol[1]) /* skip only if second chars match */
 	/* skip past all matching characters */
 	    for (; *nl && (*nl == *ol); nl++, ol++, ccs++) ;
 
 	if (!*nl) {
-	    if ((char_ins <= 0)	|| (ccs >= winw))    /* written everything */
+	    if (ccs == winw && hasam && char_ins > 0 && ins_last
+		&& vcs != winw) {
+		nl--;           /* we can assume we can go back here */
+		moveto(ln, winw - 1);
+		zputc(*nl, shout);
+		vcs++;
+		return;         /* write last character in line */
+	    }
+	    if ((char_ins <= 0) || (ccs >= winw))    /* written everything */
 		return;
-	    else		/* we've got junk on the right yet to clear */
-		if (tccan(TCCLEAREOL) && (char_ins >= tclen[TCCLEAREOL])
-	    	    && col_cleareol != -2)
-		    col_cleareol = 0;	/* force a clear to end of line */
+	    if (tccan(TCCLEAREOL) && (char_ins >= tclen[TCCLEAREOL])
+	    	&& col_cleareol != -2)
+	    /* we've got junk on the right yet to clear */
+		col_cleareol = 0;	/* force a clear to end of line */
 	}
 
 	moveto(ln, ccs);	/* move to where we do all output from */
@@ -573,7 +670,6 @@ refreshline(int ln)
     /* if we can finish quickly, do so */
 	if ((col_cleareol >= 0) && (ccs >= col_cleareol)) {
 	    tcout(TCCLEAREOL);
-	    SELECT_ADD_COST(tclen[TCCLEAREOL]);
 	    return;
 	}
 
@@ -583,10 +679,9 @@ refreshline(int ln)
 	    if (tccan(TCDEL) && (tcdelcost(i) <= i + 1))
 		tc_delchars(i);
 	    else {
-		SELECT_ADD_COST(i);
 		vcs += i;
 		while (i-- > 0)
-		    putc(' ', shout);
+		    zputc(' ', shout);
 	    }
 	    return;
 	}
@@ -595,14 +690,11 @@ refreshline(int ln)
        we can do, so we just dump out what we must and clear if we can */
 	if (!*ol) {
 	    i = (col_cleareol >= 0) ? col_cleareol : nllen;
-	    i -= ccs;
-	    fwrite(nl, i, 1, shout);
-	    SELECT_ADD_COST(i);
+	    i -= vcs;
+	    zwrite(nl, i, 1, shout);
 	    vcs += i;
-	    if (col_cleareol >= 0) {
+	    if (col_cleareol >= 0)
 		tcout(TCCLEAREOL);
-		SELECT_ADD_COST(tclen[TCCLEAREOL]);
-	    }
 	    return;
 	}
 
@@ -615,15 +707,15 @@ refreshline(int ln)
 	   eg. oldline: hifoobar \ hopefully cheaper here to delete two
 	       newline: foobar	 / characters, then we have six matches */
 	    if (tccan(TCDEL)) {
-		for (i = 1, p1 = ol + 1; *p1; p1++, i++)
-		    if (tcdelcost(i) < pfxlen(p1, nl)) {
+		for (i = 1; *(ol + i); i++)
+		    if (tcdelcost(i) < pfxlen(ol + i, nl)) {
 			tc_delchars(i);
-			SELECT_ADD_COST(i);
-			ol = p1;
+			ol += i;
 			char_ins -= i;
+			i = 0;
 			break;
 		    }
-		if (*p1)
+		if (!i)
 		    continue;
 	    }
 	/* inserting characters - characters pushed off the right should be
@@ -631,30 +723,29 @@ refreshline(int ln)
 	   undesired scrolling occurs due to `illegal' characters on screen */
 
 	    if (tccan(TCINS) && (vln != lines - 1)) {	/* not on last line */
-		for (i = 1, p1 = nl + 1; *p1; p1++, i++)
-		    if (tcinscost(i) < pfxlen(p1, ol)) {
+		for (i = 1; *(nl + i); i++)
+		    if (tcinscost(i) < pfxlen(nl + i, ol)) {
 			tc_inschars(i);
-			SELECT_ADD_COST(2 * i);
-			fwrite(nl, i, 1, shout);
-			ccs = (vcs += i);
-			nl = p1;
+			zwrite(nl, i, 1, shout);
+			nl += i;
 			char_ins += i;
+			ccs = (vcs += i);
 		    /* if we've pushed off the right, truncate oldline */
-			for (j = ccs - i, p1 = ol; *p1 && j + char_ins < winw;
-			     p1++, j++);
-			if (j + char_ins == winw)
-			    *p1 = '\0';
-			p1 = ol;
+			for (i = 0; *(ol + i) && i < winw - ccs; i++);
+			if (i == winw - ccs) {
+			    *(ol + i) = '\0';
+			    ins_last = 1;
+			}
+			i = 0;
 			break;
 		    }
-		if (*p1)
+		if (!i)
 		    continue;
 	    }
 	}
     /* we can't do any fancy tricks, so just dump the single character
        and keep on trying */
-	putc(*nl, shout);
-	SELECT_ADD_COST(1);
+	zputc(*nl, shout);
 	nl++, ol++;
 	ccs++, vcs++;
     }
@@ -670,18 +761,17 @@ moveto(int ln, int cl)
     if (vcs == winw) {
 	vln++, vcs = 0;
 	if (!hasam) {
-	    putc('\r', shout);
-	    putc('\n', shout);
+	    zputc('\r', shout);
+	    zputc('\n', shout);
 	} else {
 	    if ((vln < nlnct) && nbuf[vln] && *nbuf[vln])
-		putc(*nbuf[vln], shout);
+		zputc(*nbuf[vln], shout);
 	    else
-		putc(' ', shout);
-	    putc('\r', shout);
+		zputc(' ', shout);
+	    zputc('\r', shout);
 	    if ((vln < olnct) && obuf[vln] && *obuf[vln])
 		*obuf[vln] = *nbuf[vln];
 	}
-	SELECT_ADD_COST(2);
     }
 
     if (ln == vln && cl == vcs)
@@ -707,11 +797,9 @@ moveto(int ln, int cl)
 		vln = ln;
 		continue;
 	    }
-	putc('\r', shout), vcs = 0; /* safety precaution */
-	SELECT_ADD_COST(1);
+	zputc('\r', shout), vcs = 0; /* safety precaution */
 	while (ln > vln) {
-	    putc('\n', shout);
-	    SELECT_ADD_COST(1);
+	    zputc('\n', shout);
 	    vln++;
 	}
     }
@@ -722,8 +810,7 @@ moveto(int ln, int cl)
 /* choose cheapest movements for ttys without multiple movement capabilities -
    do this now because it's easier (to code) */
     if (cl <= vcs / 2) {
-	putc('\r', shout);
-	SELECT_ADD_COST(1);
+	zputc('\r', shout);
 	vcs = 0;
     }
     if (vcs < cl)
@@ -737,12 +824,10 @@ moveto(int ln, int cl)
 int
 tcmultout(int cap, int multcap, int ct)
 {
-    if (tccan(multcap) && (!tccan(cap) || tclen[multcap] < tclen[cap] * ct)) {
+    if (tccan(multcap) && (!tccan(cap) || tclen[multcap] <= tclen[cap] * ct)) {
 	tcoutarg(multcap, ct);
-	SELECT_ADD_COST(tclen[multcap]);
 	return 1;
     } else if (tccan(cap)) {
-	SELECT_ADD_COST((tclen[cap] * ct));
 	while (ct--)
 	    tcout(cap);
 	return 1;
@@ -754,59 +839,56 @@ tcmultout(int cap, int multcap, int ct)
 void
 tc_rightcurs(int cl)
 {
-    int ct = cl - vcs,		/* number of characters to move across	    */
-	horz_tabs = 0,		/* number of horizontal tabs if we do them  */
+    int ct,			/* number of characters to move across	    */
 	i = vcs,		/* cursor position after initial movements  */
-	j = 0;			/* number of chars outputted if we use tabs */
+	j;
     char *t;
 
-/* calculate how many horizontal tabs it would take, if we can do them -
-   tabs are assumed to be 8 spaces */
-    if (tccan(TCNEXTTAB) && ((vcs | 7) < cl)) {
-	horz_tabs = 1;
-	i = (vcs | 7) + 1;
-	for (; i + 8 <= cl; i += 8)
-	    horz_tabs++;
-	j = cl - i;		/* number of chars after last tab */
-	if (tccan(TCRIGHT))
-	    j *= tclen[TCRIGHT];
-	j += (horz_tabs * tclen[TCNEXTTAB]); /* # of chars if we use tabs */
-    }
+    ct = cl - vcs;
 
-/* do a multright if we can - if it's cheaper or we can't use other tricks */
-    if (tccan(TCMULTRIGHT) &&
-	(!tccan(TCRIGHT) || (tclen[TCMULTRIGHT] < tclen[TCRIGHT] * ct) ||
-	 !tccan(TCNEXTTAB) || (tclen[TCMULTRIGHT] < j))) {
+/* do a multright if we can - it's the most reliable */
+    if (tccan(TCMULTRIGHT)) {
 	tcoutarg(TCMULTRIGHT, ct);
-	SELECT_ADD_COST(tclen[TCMULTRIGHT]);
 	return;
     }
 
-/* try to go with tabs if a multright is not feasible/convenient */
-    if (horz_tabs) {
-	SELECT_ADD_COST((tclen[TCNEXTTAB] * horz_tabs));
-	for (; horz_tabs--;)
+/* try tabs if tabs are non destructive and multright is not possible */
+    if (!oxtabs && tccan(TCNEXTTAB) && ((vcs | 7) < cl)) {
+	i = (vcs | 7) + 1;
+	tcout(TCNEXTTAB);
+	for ( ; i + 8 <= cl; i += 8)
 	    tcout(TCNEXTTAB);
 	if ((ct = cl - i) == 0) /* number of chars still to move across */
 	    return;
     }
 
-/* or try to dump lots of right movements */
-    if (tccan(TCRIGHT)) {
-	SELECT_ADD_COST((tclen[TCRIGHT] * ct));
-	for (; ct--;)
-	    tcout(TCRIGHT);
-	return;
+/* otherwise _carefully_ write the contents of the video buffer.
+   if we're anywhere in the prompt, goto the left column and write the whole
+   prompt out unless lpptlen == pptw : we can cheat then */
+    if (vln == 0 && i < pptw) {
+	if (lpptlen == pptw)
+	    zwrite(lpptbuf + i, lpptlen - i, 1, shout);
+	else if (tccan(TCRIGHT) && (tclen[TCRIGHT] * ct <= lpptlen))
+	    /* it is cheaper to send TCRIGHT than reprint the whole prompt */
+	    for (ct = pptw - i; ct--; )
+		tcout(TCRIGHT);
+        else {
+	    if (i != 0)
+		zputc('\r', shout);
+	    zwrite(lpptbuf, lpptlen, 1, shout);
+	}
+	i = pptw;
+	ct = cl - i;
     }
 
-/* otherwise _carefully_ write the contents of the video buffer */
-    SELECT_ADD_COST(ct);
-    for (j = 0, t = nbuf[vln]; *t && (j < i); j++, t++);
-    if (j == i)
-	for ( ; *t && ct; ct--, t++)
-	    putc(*t, shout);
+    if (nbuf[vln]) {
+	for (j = 0, t = nbuf[vln]; *t && (j < i); j++, t++);
+	if (j == i)
+	    for ( ; *t && ct; ct--, t++)
+		zputc(*t, shout);
+    }
     while (ct--)
-	putc(' ', shout);	/* not my fault your terminal can't go right */
+	zputc(' ', shout);	/* not my fault your terminal can't go right */
 }
 
 /**/
@@ -816,10 +898,9 @@ tc_downcurs(int ct)
     int ret = 0;
 
     if (ct && !tcmultout(TCDOWN, TCMULTDOWN, ct)) {
-	SELECT_ADD_COST(ct + 1);
 	while (ct--)
-	    putc('\n', shout);
-	putc('\r', shout), ret = -1;
+	    zputc('\n', shout);
+	zputc('\r', shout), ret = -1;
     }
     return ret;
 }
@@ -829,13 +910,18 @@ void
 tcout(int cap)
 {
     tputs(tcstr[cap], 1, putshout);
+    SELECT_ADD_COST(tclen[cap]);
 }
 
 /**/
 void
 tcoutarg(int cap, int arg)
 {
-    tputs(tgoto(tcstr[cap], arg, arg), 1, putshout);
+    char *result;
+
+    result = tgoto(tcstr[cap], arg, arg);
+    tputs(result, 1, putshout);
+    SELECT_ADD_COST(strlen(result));
 }
 
 /**/
@@ -851,8 +937,9 @@ clearscreen(void)
 void
 redisplay(void)
 {
-    moveto(0, pptw);
-    putc('\r', shout);
+    moveto(0, 0);
+    zputc('\r', shout);		/* extra care */
+    tc_upcurs(lppth - 1);
     resetneeded = 1;
     clearflag = 0;
 }
@@ -938,7 +1025,7 @@ singlerefresh(void)
 
 	if (!*refreshop) {
 	    if ((t0 = strlen(vp)))
-		fwrite(vp, t0, 1, shout);
+		zwrite(vp, t0, 1, shout);
 	    vcs += t0;
 	    break;
 	}
@@ -947,10 +1034,10 @@ singlerefresh(void)
 		tcout(TCCLEAREOL);
 	    else
 		for (; *refreshop++; vcs++)
-		    putc(' ', shout);
+		    zputc(' ', shout);
 	    break;
 	}
-	putc(*vp, shout);
+	zputc(*vp, shout);
 	vcs++, t0++;
 	vp++, refreshop++;
     }
@@ -970,7 +1057,7 @@ singmoveto(int pos)
     if (pos == vcs)
 	return;
     if (pos <= vcs / 2) {
-	putc('\r', shout);
+	zputc('\r', shout);
 	vcs = 0;
     }
     if (pos < vcs) {
@@ -982,7 +1069,7 @@ singmoveto(int pos)
 	    vcs = pos;
 	else
 	    while (pos > vcs) {
-		putc(nbuf[0][vcs], shout);
+		zputc(nbuf[0][vcs], shout);
 		vcs++;
 	    }
     }
