@@ -51,10 +51,29 @@ spaceinline(int ct)
     int i;
 
     sizeline(ct + ll);
-    for (i = ll; i >= cs; i--)
+    for (i = ll; --i >= cs;)
 	line[i + ct] = line[i];
     ll += ct;
     line[ll] = '\0';
+
+    if (mark >= cs)
+	mark += ct;
+}
+
+/**/
+void
+shiftchars(int to, int cnt)
+{
+    if (mark >= to + cnt)
+	mark -= cnt;
+    else if (mark > to)
+	mark = to;
+
+    while (to + cnt < ll) {
+	line[to] = line[to + cnt];
+	to++;
+    }
+    ll = to;
 }
 
 /**/
@@ -64,11 +83,7 @@ backkill(int ct, int dir)
     int i = (cs -= ct);
 
     cut(i, ct, dir);
-    while (i + ct < ll) {
-	line[i] = line[i + ct];
-	i++;
-    }
-    ll -= ct;
+    shiftchars(i, ct);
 }
 
 /**/
@@ -78,11 +93,7 @@ forekill(int ct, int dir)
     int i = cs;
 
     cut(i, ct, dir);
-    while (i + ct < ll) {
-	line[i] = line[i + ct];
-	i++;
-    }
-    ll -= ct;
+    shiftchars(i, ct);
 }
 
 /**/
@@ -157,26 +168,14 @@ cut(int i, int ct, int dir)
 void
 backdel(int ct)
 {
-    int i = (cs -= ct);
-
-    while (i + ct <= ll) {
-	line[i] = line[i + ct];
-	i++;
-    }
-    ll -= ct;
+    shiftchars(cs -= ct, ct);
 }
 
 /**/
 void
 foredel(int ct)
 {
-    int i = cs;
-
-    while (i + ct <= ll) {
-	line[i] = line[i + ct];
-	i++;
-    }
-    ll -= ct;
+    shiftchars(cs, ct);
 }
 
 /**/
