@@ -52,6 +52,9 @@
  * anything                                                          */
 #undef GLOBAL_ZLOGOUT
 
+/* Define to 1 if compiler could initialise a union */
+#undef HAVE_UNION_INIT
+
 /* Define to 1 if compiler incorrectly cast signed to unsigned */
 #undef BROKEN_SIGNED_TO_UNSIGNED_CASTING
 
@@ -67,8 +70,8 @@
 /* Define to 1 if you have RFS superroot directory. */
 #undef HAVE_SUPERROOT
 
-/* Define to 1 if you have the /dev/fd filesystem */
-#undef HAVE_DEV_FD
+/* Define to the path of the /dev/fd filesystem */
+#undef PATH_DEV_FD
 
 /* Define if sys/time.h and sys/select.h cannot be both included */
 #undef TIME_H_SELECT_H_CONFLICTS
@@ -129,9 +132,6 @@
 /* Define to 1 if tgetent() accepts NULL as a buffer */
 #undef TGETENT_ACCEPTS_NULL
 
-/* Define to 1 if you have a working strcoll function */
-#undef HAVE_STRCOLL
-
 /* Define to 1 if you use POSIX style signal handling */
 #undef POSIX_SIGNALS
 
@@ -165,5 +165,14 @@
 /* Define to 1 if system has working FIFO's */
 #undef HAVE_FIFOS
 
+/* Define to 1 if struct rlimit use quad_t */
+#undef RLIM_T_IS_QUAD_T
+
+/* Define to 1 if rlimit use unsigned */
+#undef RLIM_T_IS_UNSIGNED
+
+/* Define to the type used in struct rlimit */
+#undef rlim_t
+
 /* Define to 1 if /bin/sh does not interpret \ escape sequences */
 #undef SH_USE_BSD_ECHO
